from .base_api import base_api

class twitter_recsys(base_api):
    def __init__(self):
        super().__init__()
        name = "test_twitter_recsys.parquet"
        url = f"https://pyrecdp-testdata.s3.us-west-2.amazonaws.com/{name}"            
        self.saved_path = self.download_url(name, url)

    def to_pandas(self, nrows = None):
        import pandas as pd
        return pd.read_parquet(self.saved_path)
         


from pyrecdp.core import SeriesSchema
from typing import List

class BaseFeatureGenerator:
    def __init__(self):
        pass
    
    def fit_prepare(self, pipeline, children, max_idx):
        return pipeline, children[0], max_idx


import os, sys
from pathlib import Path
try:
    import pyspark
except:
    libpath = str(Path(__file__).parent.resolve())
    spark_home_dir = f"{libpath}/spark-3.2.1-bin-hadoop3.2"
    py4j_path = f"{spark_home_dir}/python/lib/py4j-0.10.9.3-src.zip"
    os.environ["SPARK_HOME"]=spark_home_dir
    # Add the PySpark classes to the Python path:
    os.environ["PYTHONPATH"]=os.environ["SPARK_HOME"]+"/python/:"+os.environ["PYTHONPATH"]+":"+py4j_path

os.environ["JAVA_HOME"]="/usr/lib/jvm/java-8-openjdk-amd64/"
os.environ["PYSPARK_PYTHON"]=sys.executable
os.environ["PYSPARK_DRIVER_PYTHON"]=sys.executable
os.environ["PYSPARK_WORKER_PYTHON"]=sys.executable

import pyrecdp.primitives.spark_data_processor.utils as utils
sys.modules['pyrecdp.utils'] = utils
import pyrecdp.primitives.spark_data_processor.encoder as encoder
sys.modules['pyrecdp.encoder'] = encoder
import pyrecdp.primitives.spark_data_processor.data_processor as data_processor
sys.modules['pyrecdp.data_processor'] = data_processor



from .featuretools_adaptor import FeaturetoolsBasedFeatureGenerator
from pyrecdp.core import SeriesSchema
from pyrecdp.primitives.operations import Operation
from pyrecdp.core.schema import TextDtype

from featuretools.primitives.base import TransformPrimitive

class BertTokenizerDecode(TransformPrimitive):
    name = "decoded"
    return_type = TextDtype()

    def __init__(self, pretrained_model = 'bert-base-multilingual-cased', case_sensitive = False):
        self.pretrained_model = pretrained_model
        self.case_sensitive = case_sensitive

    def get_function(self):
        def decode(array):
            from transformers import BertTokenizer
            import os
            os.environ["PROTOCOL_BUFFERS_PYTHON_IMPLEMENTATION"] = "python"
            os.environ['TRANSFORMERS_OFFLINE'] = "1"
            os.environ['HF_DATASETS_OFFLINE'] = "1"
            tokenizer = BertTokenizer.from_pretrained(self.pretrained_model, do_lower_case=(not self.case_sensitive))
            
            return array.str.split().apply(lambda x: tokenizer.decode([int(n) for n in x]))

        return decode

class DecodedTextFeatureGenerator(FeaturetoolsBasedFeatureGenerator):
    def __init__(self):
        super().__init__()
        self.op_list = [
            BertTokenizerDecode
        ]
        self.op_name = 'bert_decode'

    def fit_prepare(self, pipeline, children, max_idx):
        is_useful = False
        pa_schema = pipeline[children[0]].output
        for pa_field in pa_schema:
            if pa_field.is_text and pa_field.is_encoded:
                in_feat_name = pa_field.name
                is_useful = True
                self.feature_in.append(in_feat_name)
        for in_feat_name in self.feature_in:
            self.feature_in_out_map[in_feat_name] = []
            for op in self.op_list:
                op_clz = op
                op = op_clz()
                out_feat_name = f"{in_feat_name}__{op.name}"
                out_schema = SeriesSchema(out_feat_name, op.return_type, {'is_text': True})
                self.feature_in_out_map[in_feat_name].append((out_schema.name, op_clz))
                pa_schema.append(out_schema)
        if is_useful:
            cur_idx = max_idx + 1
            config = self.feature_in_out_map
            pipeline[cur_idx] = Operation(cur_idx, children, pa_schema, op = 'bert_decode', config = config)
            return pipeline, cur_idx, cur_idx
        else:
            return pipeline, children[0], max_idx
 
class TextFeatureGenerator(FeaturetoolsBasedFeatureGenerator):
    def __init__(self):
        super().__init__()
        from featuretools.primitives import NumberOfUniqueWords, NumWords
        self.op_list = [
            NumberOfUniqueWords,
            NumWords,
        ]
        self.op_name = 'text_feature'

    def fit_prepare(self, pipeline, children, max_idx):
        pa_schema = pipeline[children[0]].output
        for pa_field in pa_schema:
            if pa_field.is_text:
                self.feature_in.append(pa_field.name)
        return super().fit_prepare(pipeline, children, max_idx)


from pyrecdp.utils import create_spark_context

import uuid
from pyrecdp.utils import *
from pyspark.ml.feature import *
import pandas as pd
import numpy as np
import pyspark.sql.types as spk_type
import pyspark.sql.types as t
import pyspark.sql.functions as spk_func
import pyspark.sql.functions as f
from pyspark.sql import *
from pyspark import *
from pyspark.ml.feature import *
import os
import sys
from timeit import default_timer as timer
import logging
import shutil
import random

class Operation:
    '''
    Base abstract class of operation

    Args:

        cols (list): columns used to generate new features or modify in place

    Attributes:

        op_name (str): operation name
        cols (list): columns
    '''

    def __init__(self, cols):
        self.op_name = "AbstractOperation"
        self.cols = cols

    def describe(self):
        return "%s(%s)" % (self.op_name, ','.join(self.cols))

    def collect(self, df):
        raise NotImplementedError(self.op_name + "doesn't support collect")

    def to_dict_dfs(self, df, spark = None, enable_gazelle=False):
        raise NotImplementedError(self.op_name + "doesn't support to_dict_dfs")

    def process(self, df, spark, df_cnt, enable_scala=True, save_path="", per_core_memory_size=0, flush_threshold = 0, enable_gazelle=False):
        return df

    # some common method
    def find_dict(self, name, dict_dfs):
        for i in dict_dfs:
            if isinstance(i, tuple):
                for v in i:
                    if isinstance(v, str):
                        dict_name = v
                    else:
                        dict_df = v
                if str(dict_name) == str(name):
                    return dict_df
            else:
                dict_df = i['dict']
                dict_name = i['col_name']
                if str(dict_name) == str(name):
                    return dict_df
        return None

    def get_colname_dict_as_tuple(self, i):
        if isinstance(i, tuple):
            for v in i:
                if isinstance(v, str):
                    dict_name = v
                else:
                    dict_df = v
            return (dict_name, dict_df)
        else:
            dict_df = i['dict']
            dict_name = i['col_name']
            return (dict_name, dict_df)

    def generate_dict_dfs(self, df, spark=None, cache_path=None, withCount=True, isParquet=True, enable_gazelle=False):
        # at least we can cache here to make read much faster
        # handle multiple columns issue
        if isParquet == False and spark != None and cache_path != None:
            cols = self.cols
            # let's raise NotImplemented exception for union case and doSplit case
            if self.doSplit:
                raise NotImplementedError("csv optimization is not supporting for doSplit yet.")
            for i in self.cols:
                if isinstance(i, dict):
                    raise NotImplementedError("csv optimization is not supporting for generate dictionary for multiple columns yet.")
            df = (df
                .select(spk_func.posexplode(spk_func.array(*cols)))
                .withColumnRenamed('pos', 'column_id')
                .withColumnRenamed('col', 'dict_col')
                .filter('dict_col is not null')
                .groupBy('column_id', 'dict_col')
                .count())
            windowed = Window.partitionBy('column_id').orderBy(spk_func.desc('count'))
            df = df.withColumn('dict_col_id', spk_func.row_number().over(windowed))
            if enable_gazelle:
                df.write.format('arrow').mode('overwrite').save(cache_path)
                df = spark.read.format("arrow").load(cache_path)
            else:
                df.write.format("parquet").mode("overwrite").save(cache_path)
                df = spark.read.parquet(cache_path)
            i = 0
            for col_name in cols:
                dict_df = df.filter('column_id == %d' % i).drop('column_id')
                if withCount:
                    self.dict_dfs.append({'col_name': col_name, 'dict': dict_df})
                else:
                    self.dict_dfs.append(
                        {'col_name': col_name, 'dict': dict_df.drop('count')})
                i += 1
            return self.dict_dfs
        # below is when input is parquet
        for i in self.cols:
            col_name = ""
            if isinstance(i, dict):
                # This is more those need to union columns
                if 'col_name' in i:
                    col_name = i['col_name']
                else:
                    col_name = 'union_dict'
                src_cols = []
                if 'src_cols' in i:
                    src_cols.extend(i['src_cols'])
                else:
                    raise ValueError(
                        "Union columns must has argument 'src_cols'")
                first = True
                dict_df = df
                for _i in src_cols:
                    if first:
                        dict_df = df.select(spk_func.col(_i).alias('dict_col'))
                        first = False
                    else:
                        dict_df = dict_df.union(
                            df.select(spk_func.col(_i).alias('dict_col')))
            else:  # if cols are simply col_name
                col_name = i
                dict_df = df.select(spk_func.col(i).alias('dict_col'))
                if self.doSplit:
                    dict_df = dict_df.select(
                        spk_func.explode(spk_func.split(spk_func.col('dict_col'), self.sep))).withColumn('dict_col', spk_func.col('col'))
            if self.bucketSize == -1:
                if self.id_by_python_dict:
                    dict_df = dict_df.distinct()
                    dict_data = dict((row['dict_col'], 1) for row in dict_df.collect())
                    for i, x in enumerate(dict_data):
                        dict_data[x] = i
                    dict_df = convert_to_spark_df(dict_data, df.spark)
                elif self.id_by_count:
                    dict_df = dict_df.groupBy('dict_col').count()
                    dict_df = dict_df.withColumn('dict_col_id', spk_func.row_number().over(
                        Window.orderBy(spk_func.desc('count')))).withColumn('dict_col_id', spk_func.col('dict_col_id') - 1).select('dict_col', 'dict_col_id', 'count')
                else:
                    dict_df = dict_df.withColumn("monotonically_increasing_id", spk_func.monotonically_increasing_id())
                    dict_df = dict_df.groupBy('dict_col').agg(spk_func.min("monotonically_increasing_id").alias("monotonically_increasing_id"), spk_func.count("*").alias("count"))
                    dict_df = dict_df.withColumn('dict_col_id', spk_func.row_number().over(
                    Window.orderBy(spk_func.col('monotonically_increasing_id')))).withColumn('dict_col_id', spk_func.col('dict_col_id') - 1).select('dict_col', 'dict_col_id', 'count')

            else:
                # when user set a bucketSize, we will quantileDiscretizer in this case
                dict_df = dict_df.groupBy('dict_col').count()
                qd = QuantileDiscretizer(numBuckets=self.bucketSize, inputCol="count", outputCol='dict_col_id')
                dict_df  = qd.fit(dict_df).transform(dict_df).withColumn('dict_col_id', spk_func.col('dict_col_id').cast(spk_type.IntegerType()))
            if withCount:
                self.dict_dfs.append({'col_name': col_name, 'dict': dict_df})
            else:
                self.dict_dfs.append(
                    {'col_name': col_name, 'dict': dict_df.drop('count')})
        return self.dict_dfs


    def categorify_strategy_decision_maker(self, dict_dfs, df, df_cnt, per_core_memory_size, flush_threshold, enable_gazelle,estimated_bytes=20):
        small_cols = []
        long_cols = []
        huge_cols = []
        smj_cols = []
        udf_cols = []
        total_small_cols_num_rows = 0
        total_estimated_shuffled_size = 0
        df_estimzate_size_per_row = sum([get_estimate_size_of_dtype(dtype) for _, dtype in df.dtypes])
        df_estimated_size = df_estimzate_size_per_row * df_cnt
        # Below threshold is maximum BHJ numRows, 4 means maximum 25% memory to cache Broadcast data, and 20 means we estimate each row has 20 bytes.
        if enable_gazelle:
            threshold = per_core_memory_size / estimated_bytes
            threshold_per_bhj = threshold if threshold <= 100000000 else 100000000
        else:
            threshold = per_core_memory_size / 4 / estimated_bytes
            threshold_per_bhj = threshold if threshold <= 30000000 else 30000000
        flush_threshold = flush_threshold * 0.8
        # print("[DEBUG] bhj total threshold is %.3f M rows, one bhj threshold is %.3f M rows, flush_threshold is %.3f GB" % (threshold / 1000000, threshold_per_bhj/1000000, flush_threshold / 2**30))
        dict_dfs_with_cnt = []
        for i in dict_dfs:
            col_name, dict_df = self.get_colname_dict_as_tuple(i)
            found = False
            if self.cols == None:
                found = True
            else:
                for j in self.cols:
                    _, col_name_from_cols = self.get_col_tgt_src(j)
                    if col_name_from_cols == col_name:
                        found = True
                        break
            if found == True:
              dict_df_cnt = dict_df.count()
              dict_dfs_with_cnt.append((col_name, dict_df, dict_df_cnt))
        sorted_dict_dfs_with_cnt = [(col_name, dict_df, dict_df_cnt) for col_name, dict_df, dict_df_cnt in sorted(dict_dfs_with_cnt, key=lambda pair: pair[2])]
        # for to_print in sorted_dict_dfs_with_cnt:
        #    print(to_print)

        sorted_cols = []
        for (col_name, dict_df, dict_df_cnt) in sorted_dict_dfs_with_cnt:
            sorted_cols.append(col_name)
            if self.doSplit:
                threshold_per_bhj = 30000000
            if (dict_df_cnt > threshold_per_bhj or (total_small_cols_num_rows + dict_df_cnt) > threshold):
                print(f"for {col_name} current dict_length {dict_df_cnt} exceeded threshold_per_bhj {threshold_per_bhj} or accumulated BHJ length {total_small_cols_num_rows + dict_df_cnt} exceeded estimated threshold {threshold}, will go smj")
                if ((total_estimated_shuffled_size + df_estimated_size) > flush_threshold):
                    # print("etstimated_to_shuffle_size for %s is %.3f GB, will do smj and spill to disk" % (str(col_name), df_estimated_size / 2**30))
                    huge_cols.append(col_name)
                    smj_cols.append(col_name)
                    long_cols.append(col_name)
                    total_estimated_shuffled_size = 0
                else:
                    # print("etstimated_to_shuffle_size for %s is %.3f GB, will do smj" % (str(col_name), df_estimated_size / 2**30))
                    smj_cols.append(col_name)
                    long_cols.append(col_name)
                    # if accumulate shuffle capacity may exceed maximum shuffle disk size, we should use hdfs instead
                    total_estimated_shuffled_size += df_estimated_size
            else:
                if self.doSplit:
                    # print("%s will do udf" % (col_name))
                    total_small_cols_num_rows += dict_df_cnt
                    udf_cols.append(col_name)
                else:
                    # print("%s will do bhj" % (col_name))
                    total_small_cols_num_rows += dict_df_cnt
                    small_cols.append(col_name)
            df_estimated_size -= df_cnt * 6
        return (sorted_cols, {'short_dict': small_cols, 'long_dict': long_cols, 'huge_dict': huge_cols, 'udf': udf_cols, 'smj_dict': smj_cols})

    def check_scala_extension(self, spark):
        driverClassPath = spark.sparkContext.getConf().get('spark.driver.extraClassPath')
        if driverClassPath == None or 'recdp' not in driverClassPath:
            return False
        else:
            return True


class FeatureModification(Operation):
    '''
    Operation to modify feature column in-place, support build-in function and udf

    Args:

        cols (list or dict): columns which are be modified
        op (str): define modify function name
        udfImpl (udf): if op is udf, udfImpl will be applied to specified cols
    '''

    def __init__(self, cols, op='udf', udfImpl=None):
        self.op_name = "FeatureModification"
        self.cols = cols
        self.op = op
        self.udf_impl = udfImpl

    def describe(self):
        if self.op == 'udf':
            f_cols = ["udf(%s)" % x for x in self.cols]
            return "%s(%s)" % (self.op_name, ','.join(f_cols))
        elif self.op == 'inline':
            f_cols = ["modify(%s, %s)" % (x, y)
                      for (x, y) in self.cols.items()]
            return "%s(%s)" % (self.op_name, ','.join(f_cols))
        else:
            f_cols = ["%s(%s)" % (self.op, x) for x in self.cols]
            return "%s(%s)" % (self.op_name, ','.join(f_cols))

    def process(self, df, spark, df_cnt, enable_scala=True, save_path="", per_core_memory_size=0, flush_threshold = 0, enable_gazelle=False):
        if self.op == 'udf':
            for i in self.cols:
                df = df.withColumn(i, self.udf_impl(spk_func.col(i)))
        elif self.op == 'inline':
            for i, func in self.cols.items():
                df = df.withColumn(i, eval(func))
        elif self.op == 'toInt':
            for i in self.cols:
                df = df.withColumn(i, spk_func.col(
                    i).cast(spk_type.IntegerType()))
        return df


class FeatureAdd(Operation):
    '''
    Operation to add new feature column based on current columns, support build-in function and udf

    Args:

        cols (dict): new_col: col or inline_func create new features
        op (str): define modify function name
        udfImpl (udf): if op is udf, udfImpl will be applied to specified cols
    '''

    def __init__(self, cols, op='udf', udfImpl=None):
        self.op_name = "FeatureAdd"
        self.cols = cols
        self.op = op
        self.udf_impl = udfImpl

    def describe(self):
        if self.op == 'udf':
            f_cols = ["new_feature(%s, udf(%s))" % (x, y)
                      for (x, y) in self.cols.items()]
            return "%s(%s)" % (self.op_name, ','.join(f_cols))
        elif self.op == 'inline':
            f_cols = ["new_feature(%s, %s)" % (x, y)
                      for (x, y) in self.cols.items()]
            return "%s(%s)" % (self.op_name, ','.join(f_cols))
        else:
            f_cols = ["new_feature(%s, %s(%s))" % (x, self.op, y)
                      for (x, y) in self.cols.items()]
            return "%s(%s)" % (self.op_name, ','.join(f_cols))

    def process(self, df, spark, df_cnt, enable_scala=True, save_path="", per_core_memory_size=0, flush_threshold = 0, enable_gazelle=False):
        if self.op == 'udf':
            for after, before in self.cols.items():
                df = df.withColumn(after, self.udf_impl(spk_func.col(before)))
        elif self.op == 'inline':
            for after, before in self.cols.items():
                try:
                    df = df.withColumn(after, eval(before))
                except Exception as e:
                    print("[ERROR]: inline script is %s" % before)
                    raise e
        elif self.op == 'toInt':
            for after, before in self.cols.items():
                df = df.withColumn(after, spk_func.col(
                    before).cast(spk_type.IntegerType()))
        return df


class FillNA(Operation):
    '''
    Operation to fillna to columns

    Args:

        cols (list): columns which are be modified
        default: filled default value
    '''

    def __init__(self, cols, default):
        self.op_name = "FillNA"
        self.cols = cols
        self.default = default

    def process(self, df, spark, df_cnt, enable_scala=True, save_path="", per_core_memory_size=0, flush_threshold = 0, enable_gazelle=False):
        return df.fillna(self.default, [i for i in self.cols])


class DropFeature(Operation):
    '''
    Operation to fillna to columns

    Args:

        cols (list): columns which are be modified
        default: filled default value
    '''

    def __init__(self, cols):
        self.op_name = "DropFeature"
        self.cols = cols

    def process(self, df, spark, df_cnt, enable_scala=True, save_path="", per_core_memory_size=0, flush_threshold = 0, enable_gazelle=False):
        for i in self.cols:
            df = df.drop(i)
        return df


class Distinct(Operation):
    '''
    Operation to perform distinct to columns

    Args:

        cols (list): columns which are be modified
        default: filled default value
    '''

    def __init__(self):
        self.op_name = "Distinct"

    def process(self, df, spark, df_cnt, enable_scala=True, save_path="", per_core_memory_size=0, flush_threshold = 0, enable_gazelle=False):
        return df.distinct()


class SelectFeature(Operation):
    '''
    Operation to select and rename columns

    Args:

        cols (list): columns which are be modified
        default: filled default value
    '''

    def __init__(self, cols=[]):
        self.op_name = "SelectFeature"
        self.cols = cols

    def process(self, df, spark, df_cnt, enable_scala=True, save_path="", per_core_memory_size=0, flush_threshold = 0, enable_gazelle=False):
        to_select = []
        for i in self.cols:
            if isinstance(i, str):
                to_select.append(i)
            elif isinstance(i, tuple):
                to_select.append("%s as %s" % (i[0], i[1]))

        return df.selectExpr(*to_select)


class Sort(Operation):
    def __init__(self, args = []):
        self.args = args

    def process(self, df, spark, df_cnt, enable_scala=True, save_path="", per_core_memory_size=0, flush_threshold = 0, enable_gazelle=False):
        return df.orderBy(*self.args)


class Categorify(Operation):
    '''
    Operation to categorify columns by distinct id

    Args:

        cols (list): columns which are be modified
    '''

    def __init__(self, cols, dict_dfs=None, gen_dicts=True, hint='auto', doSplit=False, sep='\t', doSortForArray=False, keepMostFrequent=False, saveTmpToDisk=False, multiLevelSplit=False, multiLevelSep=[],estimated_bytes=20):
        self.op_name = "Categorify"
        self.cols = cols
        self.dict_dfs = dict_dfs
        self.gen_dicts = gen_dicts
        self.hint = hint
        self.doSplit = doSplit
        self.sep = sep
        self.doSortForArray = doSortForArray
        self.keepMostFrequent = keepMostFrequent
        self.saveTmpToDisk = saveTmpToDisk
        self.save_path_id = 0
        self.multi_level_split = multiLevelSplit
        if self.multi_level_split:
            self.hint = 'udf'
        self.multi_level_sep = multiLevelSep
        self.strategy_type = {
            'udf': 'udf', 'broadcast_join': 'short_dict', 'shuffle_join': 'huge_dict'}
        self.estimated_bytes = estimated_bytes

    def process(self, df, spark, df_cnt, enable_scala=True, save_path="", per_core_memory_size=0, flush_threshold = 0, enable_gazelle=False):
        strategy = {}
        sorted_cols = []
        if self.hint != "auto" and self.hint in self.strategy_type:
            strategy[self.strategy_type[self.hint]] = []
            for i in self.dict_dfs:
                col_name, dict_df = self.get_colname_dict_as_tuple(i)
                strategy[self.strategy_type[self.hint]].append(col_name)
                sorted_cols.append(col_name)
        else:
            sorted_cols, strategy = self.categorify_strategy_decision_maker(self.dict_dfs, df, df_cnt, per_core_memory_size, flush_threshold, enable_gazelle,self.estimated_bytes)
        
        sorted_cols_pair = []
        pri_key_loaded = []
        for cn in sorted_cols:
            for i in self.cols:
                col_name, src_name = self.get_col_tgt_src(i)
                if src_name == cn:
                    if col_name not in pri_key_loaded:
                        pri_key_loaded.append(col_name)
                        sorted_cols_pair.append({col_name: src_name})


        last_method = 'bhj'
        # For now, we prepared dict_dfs and strategy
        for i in sorted_cols_pair:
            col_name, src_name = self.get_col_tgt_src(i)
            dict_df = self.find_dict(src_name, self.dict_dfs)
            if col_name != src_name and col_name not in df.columns:
                df = df.withColumn(col_name, spk_func.col(src_name))
            # for udf type, we will do udf to do user-defined categorify
            if 'udf' in strategy and src_name in strategy['udf']:
                last_method = 'udf'
                if not enable_scala and not self.multi_level_split:
                    df = self.categorify_with_udf(df, dict_df, col_name, spark)
                else:
                    df = self.categorify_with_scala_udf(df, dict_df, col_name, spark)
        for i in sorted_cols_pair:
            col_name, src_name = self.get_col_tgt_src(i)
            dict_df = self.find_dict(src_name, self.dict_dfs)
            if col_name != src_name and col_name not in df.columns:
                df = df.withColumn(col_name, spk_func.col(src_name))
            # for short dict, we will do bhj
            if 'short_dict' in strategy and src_name in strategy['short_dict']:
                last_method = 'bhj'
                df = self.categorify_with_join(df, dict_df, col_name, spark, save_path, method="bhj")
        for i in sorted_cols_pair:
            col_name, src_name = self.get_col_tgt_src(i)
            dict_df = self.find_dict(src_name, self.dict_dfs)
            if col_name != src_name and col_name not in df.columns:
                df = df.withColumn(col_name, spk_func.col(src_name))
            # for long dict, we will do shj all along
            if 'long_dict' in strategy and src_name in strategy['long_dict']:
                if 'huge_dict' in strategy and src_name in strategy['huge_dict']:
                    if 'smj_dict' in strategy and src_name in strategy['smj_dict']:
                        last_method = 'smj'
                        df = self.categorify_with_join(df, dict_df, col_name, spark, save_path, method="smj", saveTmpToDisk=True, enable_gazelle=enable_gazelle)
                    else:
                        last_method = 'shj'
                        df = self.categorify_with_join(df, dict_df, col_name, spark, save_path, method="shj", saveTmpToDisk=True, enable_gazelle=enable_gazelle)
                else:
                    if 'smj_dict' in strategy and src_name in strategy['smj_dict']:
                        last_method = 'smj'
                        df = self.categorify_with_join(df, dict_df, col_name, spark, save_path, method="smj", saveTmpToDisk=False, enable_gazelle=enable_gazelle)
                    else:
                        last_method = 'shj'
                        df = self.categorify_with_join(df, dict_df, col_name, spark, save_path, method="shj", saveTmpToDisk=False, enable_gazelle=enable_gazelle)
        # when last_method is BHJ, we should add a separator for spark wscg optimization
        if last_method == 'bhj' and not enable_gazelle and enable_scala:
            vspark = str(spark.version)
            if not (vspark.startswith("3.1") or vspark.startswith("3.0")):
                return df
            #print("Adding a CodegenSeparator to pure BHJ WSCG case")
            found = False
            for dname, dtype in df.dtypes:
                if dtype == "string":
                    df = df.withColumn(dname, spk_func.expr(f"CodegenSeparator1({dname})"))
                    found = True
                    break
                if dtype == "int":
                    df = df.withColumn(dname, spk_func.expr(f"CodegenSeparator0({dname})"))
                    found = True
                    break
                if dtype == "bigint":
                    df = df.withColumn(dname, spk_func.expr(f"CodegenSeparator2({dname})"))
                    found = True
                    break
            if found == False:
                df = df.withColumn("CodegenSeparator", spk_func.expr(f"CodegenSeparator())"))
    
        return df

    def get_col_tgt_src(self, i):
        if isinstance(i, str):
            col_name = i
            src_name = i
            return (col_name, src_name)
        elif isinstance(i, dict):
            col_name = next(iter(i.keys()))
            src_name = next(iter(i.values()))
            return (col_name, src_name)

    def get_mapping_udf(self, broadcast_data, spark, default=None):
        broadcast_dict = spark.sparkContext.broadcast(
            broadcast_data)
        sep = self.sep
        doSortForArray = self.doSortForArray
        def largest_freq_encode(x):
            broadcast_data = broadcast_dict.value
            min_val = None
            if x != '':
                x_l = x.split(sep) if not isinstance(x, list) else x
                for v in x_l:
                    if v != '' and v in broadcast_data and (min_val == None or broadcast_data[v] < min_val):
                        min_val = broadcast_data[v]
            return min_val

        def freq_encode(x):
            broadcast_data = broadcast_dict.value
            val = []
            if x != '':
                x_l = x.split(sep) if not isinstance(x, list) else x
                for v in x_l:
                    if v != '' and v in broadcast_data:
                        val.append(broadcast_data[v])
                if doSortForArray and len(val) > 0:
                    val.sort()
            return val

        if self.keepMostFrequent:
            return spk_func.udf(largest_freq_encode, spk_type.IntegerType())
        else:
            return spk_func.udf(freq_encode, spk_type.ArrayType(spk_type.IntegerType()))

    def categorify_with_udf(self, df, dict_df, i, spark):
        #print("do %s to %s" % ("python_udf", i))
        dict_data = dict((row['dict_col'], row['dict_col_id']) for row in dict_df.collect())
        udf_impl = self.get_mapping_udf(dict_data, spark)
        df = df.withColumn(i, udf_impl(spk_func.col(i)))
        return df

    def categorify_with_scala_udf(self, df, dict_df, i, spark):
        #print("do %s to %s" % ("scala_udf", i))
        # call java to broadcast data and set broadcast handler to udf
        if not self.check_scala_extension(spark):
            raise ValueError("RecDP need to enable recdp-scala-extension to run categorify_with_udf, please config spark.driver.extraClassPath and spark.executor.extraClassPath for spark")

        if self.doSplit == False:
            df = jvm_categorify(spark, df, i, dict_df)
        else:
            if get_dtype(df, i) == 'string':
                df = df.withColumn(i, spk_func.split(spk_func.col(i), self.sep))
            if self.keepMostFrequent:
                df = jvm_categorify_by_freq_for_array(spark, df, i, dict_df)
            else:
                if self.multi_level_split:
                    df = jvm_categorify_for_multi_level_array(spark, df, i, dict_df, self.multi_level_sep)
                else:
                    df = jvm_categorify_for_array(spark, df, i, dict_df)
                if self.doSortForArray:
                    df = df.withColumn(i, spk_func.array_sort(spk_func.col(i)))
        return df

    def categorify_with_join(self, df, dict_df, i, spark, save_path, method='shj', saveTmpToDisk=False, enable_gazelle=False):
        saveTmpToDisk = self.saveTmpToDisk or saveTmpToDisk
        hint_type = "shuffle_hash"
        if method == "bhj":
            hint_type = "broadcast"
        elif method == "smj":
            hint_type = "merge"
        #print("do %s to %s" % (method, i))
        to_select = df.columns + ['dict_col_id']
        if self.doSplit == False:
            df = df.join(dict_df.hint(hint_type), spk_func.col(i) == dict_df.dict_col, 'left')\
                .select(*to_select).drop(i).withColumnRenamed('dict_col_id', i)
        else:
            df = df.withColumn(
                'row_id', spk_func.monotonically_increasing_id())
            if get_dtype(df, i) == 'string':
                df = df.withColumn(i, spk_func.split(spk_func.col(i), self.sep))
            tmp_df = df.select('row_id', i).withColumn(i, spk_func.explode(spk_func.col(i)))
            tmp_df = tmp_df.join(
                dict_df.hint("shuffle_hash"),
                spk_func.col(i) == dict_df.dict_col,
                'left').filter(spk_func.col('dict_col_id').isNotNull())
            tmp_df = tmp_df.select(
                'row_id', spk_func.col('dict_col_id'))
            if self.keepMostFrequent:
                tmp_df = tmp_df.groupby('row_id').agg(spk_func.array_sort(
                    spk_func.collect_list(spk_func.col('dict_col_id'))).getItem(0).alias('dict_col_id'))
            elif self.doSortForArray:
                tmp_df = tmp_df.groupby('row_id').agg(spk_func.array_sort(
                    spk_func.collect_list(spk_func.col('dict_col_id'))).alias('dict_col_id'))
            else:
                tmp_df = tmp_df.groupby('row_id').agg(
                    spk_func.collect_list(spk_func.col('dict_col_id')).alias('dict_col_id'))

            df = df.join(tmp_df.hint(hint_type),
                         'row_id', 'left').select(*to_select).drop(i).withColumnRenamed('dict_col_id', i)
        # if self.saveTmpToDisk is True, we will save current df to hdfs or localFS instead of replying on Shuffle
        if saveTmpToDisk:
            cur_save_path = "%s_%d" % (save_path, self.save_path_id)
            self.save_path_id += 1
            if enable_gazelle:
                df.write.format('arrow').mode('overwrite').save(save_path)
                df = spark.read.format("arrow").load(save_path)
            else:
                df.write.format('parquet').mode('overwrite').save(cur_save_path)
                df = spark.read.parquet(cur_save_path)
        return df

    def categorify_with_bhj(self, df, dict_df, i, spark):
        if self.doSplit == False:
            df = df.join(dict_df.hint('broadcast'), spk_func.col(i) == dict_df.dict_col, 'left')\
                .withColumn(i, dict_df.dict_col_id).drop("dict_col_id", "dict_col")
        else:
            raise NotImplementedError(
                "We should use udf to handle withSplit + small dict scenario")
        return df


class GenerateDictionary(Operation):
    '''
    Operation to generate dictionary based on single or multi-columns by distinct id

    Args:

        cols (list): columns which are be modified
        doSplit (bool): If we need to split data
        sep (str): split seperator
    '''
    # TODO: We should add an optimization for csv input

    def __init__(self, cols, withCount=True, doSplit=False, sep='\t', isParquet=True, bucketSize=-1, id_by_count = True, id_by_python_dict = False):
        self.op_name = "GenerateDictionary"
        self.cols = cols
        self.doSplit = doSplit
        self.sep = sep
        self.withCount = withCount
        self.dict_dfs = []
        self.isParquet = isParquet
        self.bucketSize = bucketSize
        self.id_by_count = id_by_count
        self.id_by_python_dict = id_by_python_dict

    def merge_dict(self, dict_dfs, to_merge_dict_dfs):
        self.dict_dfs = []
        for i in dict_dfs:
            dict_df = i['dict']
            dict_name = i['col_name']
            to_merge = self.find_dict(dict_name, to_merge_dict_dfs)
            if to_merge == None:
                raise ValueError(
                    "Expect '%s' in merge_dicts.to_merge_dict_dfs, while find none")
            max_id_of_to_merge_row = to_merge.agg(
                {'dict_col_id': 'max'}).collect()[0]
            max_id_of_to_merge = max_id_of_to_merge_row['max(dict_col_id)']
            fwd_dict_df = dict_df.join(to_merge, 'dict_col', 'anti').withColumn('dict_col_id', spk_func.row_number().over(
                Window.orderBy('dict_col_id')) + max_id_of_to_merge)
            self.dict_dfs.append(
                {'col_name': dict_name, 'dict': to_merge.union(fwd_dict_df)})
        return self.dict_dfs

    def to_dict_dfs(self, df, spark = None, cache_path = None, enable_gazelle=False):
        return self.generate_dict_dfs(df, spark, cache_path, self.withCount, self.isParquet, enable_gazelle=enable_gazelle)


class ModelMerge(Operation):
    '''
    Operation to Merge Pre-Generated Model

    Args:
      example - dicts
      [{'col_name': ['Id'], 'dict': top_answers_df}]
    '''
    # TODO: We should add an optimization for csv input

    def __init__(self, dicts, saveTmpToDisk=False,estimated_bytes=20):
        self.op_name = "ModelMerge"
        self.dict_dfs = dicts
        self.saveTmpToDisk = saveTmpToDisk
        self.save_path_id = 0
        self.cols = None
        self.doSplit = False
        self.estimated_bytes=estimated_bytes

    def process(self, df, spark, df_cnt, enable_scala=True, save_path="", per_core_memory_size=0, flush_threshold = 0, enable_gazelle=False):
        if self.dict_dfs == None:
            raise NotImplementedError("process %s ")
        sorted_cols, strategy = self.categorify_strategy_decision_maker(self.dict_dfs, df, df_cnt, per_core_memory_size, flush_threshold, enable_gazelle,self.estimated_bytes)

        # For now, we prepared dict_dfs and strategy
        for col_name in sorted_cols:
            dict_df = self.find_dict(col_name, self.dict_dfs) 
            # for short dict, we will do bhj
            if 'short_dict' in strategy and col_name in strategy['short_dict']:
                df = self.merge_with_join(df, dict_df, col_name, spark, save_path, saveTmpToDisk=False, method='bhj', enable_gazelle=enable_gazelle)
        for col_name in sorted_cols:
            dict_df = self.find_dict(col_name, self.dict_dfs) 
            # for huge dict, we will do shj seperately
            if 'long_dict' in strategy and col_name in strategy['long_dict']:
                if 'huge_dict' in strategy and col_name in strategy['huge_dict']:
                    if 'smj_dict' in strategy and col_name in strategy['smj_dict']:
                        df = self.merge_with_join(df, dict_df, col_name, spark, save_path, saveTmpToDisk=True, method='smj', enable_gazelle=enable_gazelle)
                    else:
                        df = self.merge_with_join(df, dict_df, col_name, spark, save_path, saveTmpToDisk=True, method='shj', enable_gazelle=enable_gazelle)
                else:
                    if 'smj_dict' in strategy and col_name in strategy['smj_dict']:
                        df = self.merge_with_join(df, dict_df, col_name, spark, save_path, saveTmpToDisk=False, method='smj', enable_gazelle=enable_gazelle)
                    else:
                        df = self.merge_with_join(df, dict_df, col_name, spark, save_path, saveTmpToDisk=False, method='shj', enable_gazelle=enable_gazelle)
        return df

    def merge_with_join(self, df, dict_df, i, spark, save_path, saveTmpToDisk=False, method='shj', enable_gazelle=False):
        saveTmpToDisk = self.saveTmpToDisk or saveTmpToDisk
        hint_type = "shuffle_hash"
        if method == "bhj":
            hint_type = "broadcast"
        elif method == "smj":
            hint_type = "merge"
        #print("do %s to %s" % (method, i))
        df = df.join(dict_df.hint(hint_type), i, how='left')
        # if self.saveTmpToDisk is True, we will save current df to hdfs or localFS instead of replying on Shuffle
        if saveTmpToDisk:
            cur_save_path = "%s_%d" % (save_path, self.save_path_id)
            self.save_path_id += 1
            if enable_gazelle:
                df.write.format('arrow').mode('overwrite').save(save_path)
                df = spark.read.format("arrow").load(save_path)
            else:
                df.write.format('parquet').mode('overwrite').save(cur_save_path)
                df = spark.read.parquet(cur_save_path)
        return df


class NegativeSample(Operation):
    def __init__(self, cols, dicts, negCnt = 1):
        self.op_name = "NegativeSample"
        self.cols = cols
        self.dict_dfs = dicts
        self.neg_cnt = negCnt
        if len(cols) != len(dicts):
            raise ValueError("NegativeSample expects input dicts has same size cols for mapping")
        self.num_cols = len(cols)

    def get_negative_sample_udf(self, spark, broadcast_data):
        broadcast_movie_id_list = spark.sparkContext.broadcast(
            broadcast_data)

        def get_random_id(asin):
            item_list = broadcast_movie_id_list.value
            asin_total_len = len(item_list)
            asin_neg = asin
            while True:
                asin_neg_index = random.randint(0, asin_total_len - 1)
                asin_neg = item_list[asin_neg_index]
                if asin_neg == None or asin_neg == asin:
                    continue
                else:
                    break
            return [asin_neg, asin]

        return spk_func.udf(get_random_id, spk_type.ArrayType(spk_type.StringType()))


    def process(self, df, spark, df_cnt, enable_scala=True, save_path="", per_core_memory_size=0, flush_threshold = 0, enable_gazelle=False):
        if enable_scala:
            for i in range(0, self.num_cols):
                df = jvm_get_negative_sample(spark, df, self.cols[i], self.get_colname_dict_as_tuple(self.dict_dfs[i])[1], self.neg_cnt)
        else:
            udf_list = []
            for i in range(0, self.num_cols):
                broadcasted_data = [row['dict_col'] for row in self.get_colname_dict_as_tuple(self.dict_dfs[i])[1].select("dict_col").collect()]
                negative_sample = self.get_negative_sample_udf(spark, broadcasted_data)
                udf_list.append(negative_sample)
            for i in range(0, self.num_cols):
                df = df.withColumn(self.cols[i], udf_list[i](spk_func.col(self.cols[i])))
                df = df.select(spk_func.col("*"), spk_func.posexplode(spk_func.col(self.cols[i]))).drop(self.cols[i]).withColumnRenamed("col", self.cols[i])
        return df


class NegativeFeature(Operation):
    def __init__(self, cols, dicts, doSplit=False, sep='\t', negCnt = 1):
        self.op_name = "NegativeFeature"
        self.cols = cols
        self.dict_dfs = dicts
        self.doSplit = doSplit
        self.neg_cnt = negCnt
        self.sep = sep
        if len(cols) != len(dicts):
            raise ValueError("NegativeFeature expects input dicts has same size cols for mapping")
        self.num_cols = len(cols)

    def get_negative_feature_udf(self, spark, broadcast_data):
        broadcast_movie_id_list = spark.sparkContext.broadcast(
            broadcast_data)

        if self.doSplit:
            def get_random_id(hist_asin):
                item_list = broadcast_movie_id_list.value
                asin_total_len = len(item_list)
                res = []
                asin_list = hist_asin.split(self.sep) if not isinstance(hist_asin, list) else hist_asin
                for asin in asin_list:
                    asin_neg = asin
                    while True:
                        asin_neg_index = random.randint(0, asin_total_len - 1)
                        asin_neg = item_list[asin_neg_index]
                        if asin_neg == None or asin_neg == asin:
                            continue
                        else:
                            res.append(asin_neg)
                            break
                return res

            return spk_func.udf(get_random_id, spk_type.ArrayType(spk_type.StringType()))
        else:
            def get_random_id(asin):
                item_list = broadcast_movie_id_list.value
                asin_total_len = len(item_list)
                asin_neg = asin
                while True:
                    asin_neg_index = random.randint(0, asin_total_len - 1)
                    asin_neg = item_list[asin_neg_index]
                    if asin_neg == None or asin_neg == asin:
                        continue
                    else:
                        break
                return asin_neg

            return spk_func.udf(get_random_id, spk_type.StringType())


    def process(self, df, spark, df_cnt, enable_scala=True, save_path="", per_core_memory_size=0, flush_threshold = 0, enable_gazelle=False):
        if enable_scala:
            for tgt, src in self.cols.items():
                dict_df = self.find_dict(src, self.dict_dfs)
                if self.doSplit:
                    src_type = get_dtype(df, src)
                    if src_type == 'string':
                        df = df.withColumn(tgt, spk_func.split(spk_func.col(src), self.sep))
                        df = jvm_get_negative_feature_for_array(spark, df, tgt, tgt, dict_df, self.neg_cnt)
                    else:
                        df = jvm_get_negative_feature_for_array(spark, df, tgt, src, dict_df, self.neg_cnt)
                else:
                    df = jvm_get_negative_feature(spark, df, tgt, src, dict_df, self.neg_cnt)
        else:
            for tgt, src in self.cols.items():
                broadcasted_data = [row['dict_col'] for row in self.find_dict(src, self.dict_dfs)[1].select("dict_col").collect()]
                negative_feature = self.get_negative_feature_udf(spark, broadcasted_data)
                df = df.withColumn(tgt, negative_feature(spk_func.col(src)))
        return df


class ScalaDFTest(Operation):
    def __init__(self, cols, dicts=None, method="int"):
        self.op_name = "ScalaDFTest"
        self.cols = cols
        self.dict_dfs = dicts
        if dicts != None and len(cols) != len(dicts):
            raise ValueError("ScalaDFTest expects input dicts has same size cols for mapping")
        self.num_cols = len(cols)
        self.method = method


    def process(self, df, spark, df_cnt, enable_scala=True, save_path="", per_core_memory_size=0, flush_threshold = 0, enable_gazelle=False):
        if self.method == "int":
            for i in range(0, self.num_cols):
                df = jvm_scala_df_test_int(spark, df, self.cols[i])
        elif self.method == "str":
            for i in range(0, self.num_cols):
                df = jvm_scala_df_test_str(spark, df, self.cols[i])
        elif self.method == "broadcast" and self.dict_dfs != None:
            for i in range(0, self.num_cols):
                df = jvm_scala_df_test_broadcast(spark, df, self.cols[i], self.get_colname_dict_as_tuple(self.dict_dfs[i])[1])
        return df


class CollapseByHist(Operation):
    def __init__(self, cols = [], by = None, orderBy = None, minNumHist = 0, maxNumHist = -1):
        self.op_name = "CollapseByHist"
        self.cols = cols
        self.by = by
        self.orderBy = orderBy
        self.minNumHist = minNumHist
        self.maxNumHist = maxNumHist
        if by == None or (isinstance(by, list) and len(by) == 0):
            raise ValueError("CollapseByHist expects input by should not None or Empty")


    def process(self, df, spark, df_cnt, enable_scala=True, save_path="", per_core_memory_size=0, flush_threshold = 0, enable_gazelle=False):
        w = Window.partitionBy(self.by)
        last_collapse_col = ""
        if self.orderBy != None:
            w = w.orderBy(self.orderBy)
        df = df.withColumn('row_id', spk_func.row_number().over(w))
        for c in self.cols:
            df = df.withColumn(f'hist_{c}', spk_func.collect_list(c).over(Window.partitionBy(self.by)))
            df = df.withColumn(f'hist_{c}', spk_func.when(spk_func.col(f'hist_{c}').isNull(), spk_func.array()).otherwise(spk_func.col(f'hist_{c}')))
            last_collapse_col = f'hist_{c}'
        if last_collapse_col == "":
            if self.orderBy != None:
                df = df.withColumn('row_max', spk_func.max(self.orderBy).over(Window.partitionBy(self.by)))
                df = df.filter((df[self.orderBy] == df.row_max))
                df = df.drop('row_id').drop('row_max')
            else:
                df = df.withColumn('row_cnt', spk_func.max('row_id').over(Window.partitionBy(self.by)))
                df = df.filter((df.row_id == df.row_cnt) & (df.row_cnt > spk_func.lit(self.minNumHist)))
                df = df.drop('row_id')
        else:
            df = df.withColumn('row_cnt', spk_func.size(spk_func.col(last_collapse_col)))
            df = df.withColumn('row_cnt', spk_func.col('row_cnt').cast(spk_type.IntegerType()))
            df = df.filter((df.row_id == df.row_cnt) & (df.row_cnt > spk_func.lit(self.minNumHist)))
            for c in self.cols:
                if self.maxNumHist != -1:
                    df = df.withColumn('max_hist_len', spk_func.when(spk_func.col('row_cnt') > self.maxNumHist, self.maxNumHist).otherwise(spk_func.col('row_cnt')))
                else:
                    df = df.withColumn('max_hist_len', spk_func.col('row_cnt'))
                df = df.withColumn(f'hist_{c}', spk_func.expr(f"slice(hist_{c}, 1, max_hist_len - 1)"))
            df = df.drop('row_id').drop('row_cnt').drop('max_hist_len')
        return df


class DataProcessor:
    def __init__(self, spark = None, path_prefix="file://", current_path="", shuffle_disk_capacity="unlimited", dicts_path="dicts", spark_mode='local', enable_gazelle=False):
        self.ops = []
        self.stop_spark_in_del = False
        if not spark:
            spark = create_spark_context()
            self.stop_spark_in_del = True
        self.spark = spark
        self.uuid = uuid.uuid1()
        self.tmp_id = 0
        self.path_prefix = path_prefix
        self.current_path = current_path
        self.dicts_path = dicts_path
        self.tmp_materialzed_list = []
        self.per_core_memory_size = 0
        self.enable_gazelle = enable_gazelle
        self.spark_mode = spark_mode
        fail_to_parse = True
        if spark_mode == 'yarn' or spark_mode == 'standalone':
            try:
                memory_size = parse_size(spark.sparkContext.getConf().get('spark.executor.memory'))
                memory_size += parse_size(spark.sparkContext.getConf().get('spark.executor.memoryOverhead'))
                numCores = int(spark.sparkContext.getConf().get('spark.executor.cores'))
                self.per_core_memory_size = memory_size / numCores
                fail_to_parse = False
            except:
                pass
        elif spark_mode == 'local' or fail_to_parse:
            memory_size = parse_size(spark.sparkContext.getConf().get('spark.driver.memory'))
            numCores = int(parse_cores_num(spark.sparkContext.getConf().get('spark.master')))
            self.per_core_memory_size = memory_size / numCores
            self.spark_mode = 'local'
        if shuffle_disk_capacity == "unlimited":
            self.flush_threshold = (2**63 - 1)
        else:
            self.flush_threshold = parse_size(shuffle_disk_capacity)
        self.gateway = spark.sparkContext._gateway
        self.enable_scala = self.registerScalaUDFs()
        print("per core memory size is %.3f GB and shuffle_disk maximum capacity is %.3f GB" % (self.per_core_memory_size * 1.0/(2**30), self.flush_threshold * 1.0/(2**30)))

    def __del__(self):
        if self.stop_spark_in_del:
            self.spark.stop()
        for tmp_file in self.tmp_materialzed_list:
            shutil.rmtree(tmp_file, ignore_errors=True)

    def registerScalaUDFs(self):
        driverClassPath = self.spark.sparkContext.getConf().get('spark.driver.extraClassPath')
        execClassPath = self.spark.sparkContext.getConf().get('spark.executor.extraClassPath')
        if self.spark_mode == 'local':
            if driverClassPath == None or 'recdp' not in driverClassPath:
                return False
        else: 
            if driverClassPath == None or execClassPath == None or 'recdp' not in driverClassPath or 'recdp' not in execClassPath:
                return False
        print("recdp-scala-extension is enabled")
        self.spark.udf.registerJavaFunction("sortStringArrayByFrequency","com.intel.recdp.SortStringArrayByFrequency")
        self.spark.udf.registerJavaFunction("sortIntArrayByFrequency","com.intel.recdp.SortIntArrayByFrequency")
        self.spark._jsparkSession.udf().register("CodegenSeparator", self.gateway.jvm.org.apache.spark.sql.api.CodegenSeparator())
        self.spark._jsparkSession.udf().register("CodegenSeparator0", self.gateway.jvm.org.apache.spark.sql.api.CodegenSeparator0())
        self.spark._jsparkSession.udf().register("CodegenSeparator1", self.gateway.jvm.org.apache.spark.sql.api.CodegenSeparator1())
        self.spark._jsparkSession.udf().register("CodegenSeparator2", self.gateway.jvm.org.apache.spark.sql.api.CodegenSeparator2())
        return True

    def describe(self):
        description = []
        for op in self.ops:
            description.append(op.describe())
        print("DataProcessor current worflow is \n    ",
              '\n    ->'.join(description))

    def append_ops(self, ops):
        self.ops += ops

    def reset_ops(self, ops):
        self.ops = ops

    def get_tmp_cache_path(self):
        tmp_id = self.tmp_id
        self.tmp_id += 1
        save_path = "%s/%s/tmp/%s-%s-%d" % (
            self.path_prefix, self.current_path, "materialized_tmp", self.uuid, tmp_id)
        self.tmp_materialzed_list.append(save_path)
        return save_path

    def materialize(self, df, df_name="materialized_tmp", method=1):
        if method == 0:
            return df.cache()
        else:
            save_path = ""
            if df_name == "materialized_tmp":
                save_path = self.get_tmp_cache_path()
                self.tmp_materialzed_list.append(save_path)
            else:
                save_path = "%s/%s/%s" % (self.path_prefix,
                                          self.current_path, df_name)
            print(f"save data to {save_path}")
            if self.enable_gazelle:
                #df.write.format('parquet').mode('overwrite').save(save_path)
                df.write.format('arrow').mode('overwrite').save(save_path)
                return self.spark.read.format("arrow").load(save_path)
            else:
                df.write.format('parquet').mode('overwrite').save(save_path)
                return self.spark.read.parquet(save_path)

    def refine_op(self, op, df, save_path):
        if isinstance(op, Categorify) and op.dict_dfs == None :
            dict_dfs = []
            if not op.gen_dicts:
                dict_names = op.cols
                dict_dfs = [{'col_name': name, 'dict': self.spark.read.parquet(
                    "%s/%s/%s/%s" % (self.path_prefix, self.current_path, self.dicts_path, name))} for name in dict_names]
            if op.gen_dicts:
                dfs = []
                op_gen_dicts = GenerateDictionary(op.cols, doSplit=op.doSplit, sep=op.sep)
                dfs.extend(op_gen_dicts.to_dict_dfs(df, self.spark, save_path, self.enable_gazelle))
                for dict_df in dfs:
                    dict_dfs.append({'col_name': dict_df['col_name'], 'dict': self.materialize(
                        dict_df['dict'], "%s/%s" % (self.dicts_path, dict_df['col_name']))})
            op.dict_dfs = dict_dfs
        return op

    def apply(self, df, df_cnt = None):
        if len(self.ops) > 0 and df_cnt == None:
            df_cnt = df.count()
        for op in self.ops:
            save_path = self.get_tmp_cache_path()
            op = self.refine_op(op, df, save_path)
            df = op.process(df, self.spark, df_cnt, self.enable_scala, save_path=save_path, per_core_memory_size = self.per_core_memory_size, flush_threshold = self.flush_threshold, enable_gazelle=self.enable_gazelle)
        self.ops = []
        return df

    def transform(self, df, name="materialized_tmp", df_cnt = None):
        return self.materialize(self.apply(df, df_cnt), df_name=name)

    def generate_dicts(self, df):
        # flat ops to dfs
        dfs = []
        materialized_dfs = []
        for op in self.ops:
            if op.op_name != "GenerateDictionary":
                raise NotImplementedError(
                    "We haven't support apply generate_dict to not GenerateDictionary operator yet.")
            save_path = self.get_tmp_cache_path()
            dfs.extend(op.to_dict_dfs(df, self.spark, save_path, self.enable_gazelle))
        for dict_df in dfs:
            materialized_dfs.append({'col_name': dict_df['col_name'], 'dict': self.materialize(
                dict_df['dict'], "%s/%s" % (self.dicts_path, dict_df['col_name']))})
        return materialized_dfs

    def merge_dicts(self, df, to_merge_dict_dfs):
        # flat ops to dfs
        dfs = []
        materialized_dfs = []
        for op in self.ops:
            if op.op_name != "GenerateDictionary":
                raise NotImplementedError(
                    "We haven't support apply generate_dict to not GenerateDictionary operator yet.")
            save_path = get_tmp_cache_path()
            dfs.extend(op.merge_dict(op.to_dict_dfs(df, self.spark, save_path, self.enable_gazelle), to_merge_dict_dfs))
        for dict_df in dfs:
            materialized_dfs.append({'col_name': dict_df['col_name'], 'dict': self.materialize(
                dict_df['dict'], "%s/%s_merged" % (self.dicts_path, dict_df['col_name']))})
        return materialized_dfs

    def get_sample(self, df, df_cnt = None, vertical = False, truncate = 50):
        if df_cnt == None:
            df_cnt = df.count()
        for op in self.ops:
            save_path = self.get_tmp_cache_path()
            op = self.refine_op(op, df, save_path)
            df = op.process(df, self.spark, df_cnt, self.enable_scala, save_path=save_path, per_core_memory_size = self.per_core_memory_size, flush_threshold = self.flush_threshold, enable_gazelle=self.enable_gazelle)
        df.show(vertical = vertical, truncate = truncate)



from .base_api import base_api

class amazon_product_review(base_api):
    def __init__(self):
        super().__init__()
        name = "amazon_reviews_us_Books.tsv"
        url = "https://s3.amazonaws.com/amazon-reviews-pds/tsv/amazon_reviews_us_Books_v1_00.tsv.gz"            
        self.saved_path = self.download_url(name, url, unzip = True)

    def to_pandas(self, nrows = None):
        import pandas as pd
        df = pd.read_table(self.saved_path, on_bad_lines='skip')
        
        # fix train
        df = df.loc[df['star_rating'].apply(lambda x: len(str(x)) <= 3)]
        df['star_rating'] = df['star_rating'].astype(float)
        
        return df
         


from pyrecdp.primitives.generators import *
from pyrecdp.core import DataFrameSchema, SparkDataProcessor, DiGraph
from pyrecdp.primitives.operations import Operation, DataFrameOperation, RDDToDataFrameConverter
import pandas as pd
import logging
import graphviz
import json
from pyrecdp.core.utils import Timer, sample_read, deepcopy
from pyspark.sql import DataFrame as SparkDataFrame
from pyspark import RDD as SparkRDD
from IPython.display import display

logging.basicConfig(format='%(asctime)s %(levelname)s:%(message)s', level=logging.ERROR, datefmt='%I:%M:%S')
logger = logging.getLogger(__name__)

class BasePipeline:
    def __init__(self, dataset, label, *args, **kwargs):
        if isinstance(dataset, pd.DataFrame):
            self.dataset = {'main_table': dataset}
        elif isinstance(dataset, list):
            self.dataset = dict((idx, data) for idx, data in enumerate(dataset))
        elif isinstance(dataset, dict):
            self.dataset = dataset
        else:
            self.dataset = {'main_table': dataset}
        main_table = None
        input_is_path = False
        if isinstance(label, str):    
            for data_key, data in self.dataset.items():
                # optional: data is file_path
                if isinstance(data, str):
                    input_is_path = True
                    data = sample_read(data)
                if label in data.columns:
                    main_table = data_key
                    break
            self.y = label
        elif isinstance(label, pd.Series):
            self.y = label.name
        else:
            self.y = None
        if not main_table:
            raise ValueError(f"label {label} is not found in dataset")
        self.main_table = main_table
        
        # Set properties for BasePipeline
        if not input_is_path:
            original_data = self.dataset[main_table]
        else:
            original_data = sample_read(self.dataset[main_table])
        
        self.feature_columns = [i for i in original_data.columns if i != self.y]
        feature_data = original_data[self.feature_columns]
            
        self.generators = []
        self.pipeline = DiGraph()
        if not input_is_path:
            op = 'DataFrame'
            config = main_table
        else:
            op = 'DataLoader'
            config = {'table_name': main_table, 'file_path': self.dataset[main_table]}
        
        cur_id = 0
        self.pipeline[cur_id] = Operation(
            cur_id, None, output = DataFrameSchema(feature_data), op = op, config = config)

        if len(self.dataset) > 1:
            self.supplementary = dict((k, v) for k, v in self.dataset.items() if k != main_table)
        else:
            self.supplementary = None
        self.rdp = None
     
    def fit_analyze(self, *args, **kwargs):
        child = list(self.pipeline.keys())[-1]
        max_id = child
        for i in range(len(self.generators)):
            for generator in self.generators[i]:
                self.pipeline, child, max_id = generator.fit_prepare(self.pipeline, [child], max_id)
        return child, max_id

    def __repr__(self):
        return repr(self.pipeline)

    def export(self, file_path = None):
        json_object = self.pipeline.json_dump()
        if file_path:
            # Writing to sample.json
            with open(file_path, "w") as outfile:
                outfile.write(json_object)
        else:
            print(json_object)
    
    def import_from_json(self, file_path):
        with open(file_path, "r") as infile:
            json_object = json.load(infile)
        for idx, op_config in json_object.items():
            idx = int(idx)
            if idx in self.pipeline:
                continue
            self.pipeline[idx] = Operation.load(idx, op_config)
        #self.create_executable_pipeline()
        
    def create_executable_pipeline(self):
        node_chain = self.pipeline.convert_to_node_chain()
        executable_pipeline = DiGraph()
        executable_sequence = []
        for idx in node_chain:
            executable_pipeline[idx] = self.pipeline[idx].instantiate()
            executable_sequence.append(executable_pipeline[idx])
        return executable_pipeline, executable_sequence
    
    def add_operation(self, config):
        # check if below keys are existing
        # children: who is its child
        # next: who is will use this operation
        # inline_function: the function to process data
        example = """
            def add_one_print(df):
                df['col_1'] = df['col_1'] + 1
                display(df)
                return df
            {
                "children": [7],
                "next": [8],
                "inline_function": add_one_print,
            }"""
        try:
            children = config["children"]
            next = config["next"]
            inline_function = config["inline_function"]
        except:
            raise ValueError(f"please follow this example to create customer operator: {example}")
        if not isinstance(children, list):
            children = [children]
        if not isinstance(next, list):
            next = [next]
        if len(children) > 1 and len(next) > 1:
            raise NotImplementedError("Current we didn't support add operation to case that children and next are all multiple operators.")
        
        # get current max operator id
        max_idx = self.pipeline.get_max_idx()
        cur_idx = max_idx + 1
        
        config = {
            "func_name": inline_function,
        }
        self.pipeline[cur_idx] = Operation(
            cur_idx, children, output = None, op = "custom_operator", config = config)
        for idx in next:
            # replace next's children with new added operator
            children_in_next = self.pipeline[idx].children
            found = {}
            for id, child in enumerate(children_in_next):
                if child in children:
                    found[id] = cur_idx
            for k, v in found.items():
                self.pipeline[idx].children[k] = v
        display(self.plot())

    def delete_operation(self, config):
        example = """
            config = {
                "idx": 7,
                "next": [8]
            }"""
        try:
            cur_idx = config["idx"]
            next = config["next"]
        except:
            raise ValueError(f"please follow this example to create customer operator: {example}")
        if not isinstance(next, list):
            next = [next]
            
        children = self.pipeline[cur_idx].children
        if len(children) > 1 and len(next) > 1:
            raise NotImplementedError("Current we didn't support add operation to case that children and next are all multiple operators.")
        
        if len(next) == 1:
            self.pipeline[next[0]].children = children
        else:            
            for idx in next:
                # replace next's children with new added operator
                children_in_next = self.pipeline[idx].children
                found = {}
                for id, child in enumerate(children_in_next):
                    if child == cur_idx:
                        found[id] = children[0]
                for k, v in found.items():
                    self.pipeline[idx].children[k] = v
        del self.pipeline[cur_idx]
        display(self.plot())        
           
    def plot(self):
        f = graphviz.Digraph()
        edges = []
        nodes = []
        f.attr(fontsize='10')
        def add_escape(input):
            input = input.replace('<', '\<').replace('>', '\>')
            #input = input.replace("'", "\\\'").replace("\"", "\\\"")
            return input

        def add_break(input):
            if isinstance(input, list) and len(input) < 3:
                for line in input:
                    if isinstance(line, str):
                        ret = str(input)
                        return ret
            if isinstance(input, dict):
                input = [f"{k}: {add_break(v)}" for k, v in input.items()]
            if isinstance(input, list):
                try_str = str(input)
                if len(try_str) < 200:
                    return try_str
                ret = "" + "\l"
                for line in input:
                    ret += str(add_break(line)) + "\l"
                return ret
            return input

        for node_id, config in self.pipeline.items():
            nodes.append([str(node_id), f"{node_id}:{config.op} |{add_escape(str(add_break(config.config)))}"])
            if config.children:
                for src_id in config.children:
                    edges.append([str(src_id), str(node_id)])
        for node in nodes:
            f.node(node[0], node[1], shape='record', fontsize='12')
        for edge in edges:
            f.edge(*edge)
        return f  

    def to_chain(self):
        return self.pipeline.convert_to_node_chain()
        
    def execute(self, engine_type = "pandas", start_op_idx = 0, no_cache = False, transformed_end = -1):
        # prepare pipeline
        executable_pipeline, executable_sequence = self.create_executable_pipeline()
        print(executable_pipeline)

        # execute
        if engine_type == 'pandas':
            with Timer(f"execute with pandas"):
                start = False
                for op in executable_sequence:
                    if op.op.idx == start_op_idx:
                        start = True
                    if not start:
                        continue
                    if isinstance(op, DataFrameOperation):
                        input_df = self.dataset if start_op_idx == 0 else {'main_table': self.transformed_cache}
                        input_df = deepcopy(input_df) if no_cache else input_df
                        op.set(input_df)
                    with Timer(f"execute {op}"):
                        op.execute_pd(executable_pipeline, no_cache)
            if transformed_end == -1:
                df = executable_sequence[-1].cache
            else:
                df = executable_pipeline[transformed_end].cache
        elif engine_type == 'spark':
            with Timer(f"execute with spark"):
                start = False
                for op in executable_sequence:
                    if op.op.idx == start_op_idx:
                        start = True
                    if not start:
                        continue
                    if isinstance(op, DataFrameOperation) and op.op.idx == start_op_idx:
                        input_df = self.dataset if start_op_idx == 0 else {'main_table': self.transformed_cache}
                        input_df = deepcopy(input_df) if no_cache else input_df
                        op.set(input_df)
                    print(f"append {op}")
                    op.execute_spark(executable_pipeline, self.rdp, no_cache)
                if transformed_end == -1:
                    ret = executable_sequence[-1].cache
                else:
                    ret = executable_pipeline[transformed_end].cache
                if isinstance(ret, SparkDataFrame):
                    df = self.rdp.transform(ret)
                elif isinstance(ret, SparkRDD):
                    _convert = RDDToDataFrameConverter().get_function(self.rdp)
                    df = _convert(ret)
                else:
                    df = ret
        else:
            raise NotImplementedError('pipeline only support pandas and spark as engine')
        
        # fetch result
        return df

    def fit_transform(self, engine_type = 'pandas', no_cache = False, *args, **kwargs):
        if not no_cache and hasattr(self, 'transformed_cache') and self.transformed_cache is not None:
            print("Detect pre-transformed cache, return cached data")
            print("If re-transform is required, please use fit_transform(no_cache = True)")
            return self.transformed_cache
        if engine_type == "spark":
            self.rdp = SparkDataProcessor()
        ret = self.execute(engine_type, no_cache)
        if engine_type == "spark":
            del self.rdp 
            self.rdp = None
        self.transformed_cache = ret
        return ret
    
    def get_transformed_cache(self):
        if hasattr(self, 'transformed_cache') and self.transformed_cache is not None:
            return self.transformed_cache
        else:
            print("No transformed data detected.")


from .base import BaseFeatureGenerator as super_class
import pandas as pd
from pyrecdp.core import SeriesSchema
from pyrecdp.primitives.operations import Operation

class LabelEncodeFeatureGenerator(super_class):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def fit_prepare(self, pipeline, children, max_idx):
        is_useful = False
        pa_schema = pipeline[children[0]].output
        feature_in_out = {}
        folder = 'pipeline_default'
        for idx, pa_field in enumerate(pa_schema):
            if pa_field.is_string:
                feature = pa_field.name
                out_schema = SeriesSchema(f"{feature}", pd.CategoricalDtype())
                feature_in_out[feature] = (f"{folder}/{feature}_categorify_dict", feature)
                is_useful = True
                pa_schema[idx] = out_schema
        if is_useful:
            cur_idx = max_idx + 1
            config = feature_in_out
            pipeline[cur_idx] = Operation(cur_idx, children, pa_schema, op = 'categorify', config = config)
            return pipeline, cur_idx, cur_idx
        else:
            return pipeline, children[0], max_idx

class OneHotFeatureGenerator(super_class):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.feature_in = []

    def fit_prepare(self, pipeline, children, max_idx):
        is_useful = False
        pa_schema = pipeline[children[0]].output
        config = {}
        for pa_field in pa_schema:
            if pa_field.is_onehot:
                feature = pa_field.name
                out_schema = [SeriesSchema(f"{feature}__{key}", int) for key in pa_field.config["is_onehot"]]
                config[pa_field.name] = pa_field.config["is_onehot"]
                is_useful = True
                pa_schema.extend(out_schema)
        if is_useful:
            cur_idx = max_idx + 1
            pipeline[cur_idx] = Operation(cur_idx, children, pa_schema, op = 'onehot_encode', config = config)
            return pipeline, cur_idx, cur_idx
        else:
            return pipeline, children[0], max_idx
        
class ListOneHotFeatureGenerator(super_class):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.feature_in = []

    def fit_prepare(self, pipeline, children, max_idx):
        is_useful = False
        pa_schema = pipeline[children[0]].output
        config = {}
        for pa_field in pa_schema:
            if pa_field.is_list_string:
                feature = pa_field.name
                out_schema = [SeriesSchema(f"{feature}_{key}", int) for key in pa_field.config["is_list_string"][1] if key is not None or key is not ""]
                config[pa_field.name] = pa_field.config["is_list_string"]
                is_useful = True
                pa_schema.extend(out_schema)
        if is_useful:
            cur_idx = max_idx + 1
            pipeline[cur_idx] = Operation(cur_idx, children, pa_schema, op = 'list_onehot_encode', config = config)
            return pipeline, cur_idx, cur_idx
        else:
            return pipeline, children[0], max_idx
        
class TargetEncodeFeatureGenerator(super_class):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.feature_in = []

    def fit_prepare(self, pipeline, children, max_idx):
        is_useful = False
        pa_schema = pipeline[children[0]].output
        config = {}
        for pa_field in pa_schema:
            if pa_field.is_onehot:
                feature = pa_field.name
                out_schema = [SeriesSchema(f"{feature}__{key}", int) for key in pa_field.config["is_onehot"]]
                config[pa_field.name] = pa_field.config["is_onehot"]
                is_useful = True
                pa_schema.extend(out_schema)
        if is_useful:
            cur_idx = max_idx + 1
            pipeline[cur_idx] = Operation(cur_idx, children, pa_schema, op = 'categorify', config = config)
            return pipeline, cur_idx, cur_idx
        else:
            return pipeline, children[0], max_idx


from .base_api import base_api

class ibm_fraud_detect(base_api):
    def __init__(self, scale = 'full'):
        super().__init__()
        if scale == 'test':
            raise NotImplementedError("ibm_fraud_detect test dataset is not created yet")
        else:
            url = "https://huggingface.co/datasets/Chendi/ibm_transactions/resolve/main/transactions.tgz"         
            self.saved_path = self.download_url("card_transaction.v1.csv", url, unzip = True)

    def to_pandas(self, scale = 'full'):
        import pandas as pd
        if scale == 'test':
            return pd.read_csv(self.saved_path, nrows = 100000)
        return pd.read_csv(self.saved_path)
         


from woodwork.column_schema import ColumnSchema
from pandas import StringDtype
from pyrecdp.core.utils import is_text_series, is_tuple, is_encoded, is_integer_convertable
class TextDtype(StringDtype):
    pass

class SeriesSchema:
    def __init__(self, *args):
        if len(args) == 1:
            s = args[0]
            self.name = s.name
            in_type = s.dtype
            self.config = {}
            self.config['is_text'] = is_text_series(s)
            if self.config['is_text']:
                self.config['is_encoded'] = is_encoded(s)
            self.config['is_tuple'] = is_tuple(s)
            self.config['is_integer'] = is_integer_convertable(s)
        elif len(args) >= 2:
            # s_dtype is possible to be pandas.dtype or woodwork.dtype       
            self.name = args[0]
            # TODO: convert featuretools return_type to recdp return type
            in_type = args[1]
            if len(args) > 2:
                self.config = args[2]
            else:
                self.config = {}
            self.config['is_integer'] = False
        else:
            raise ValueError(f"SeriesSchema unsupport input as {args}")

        if in_type:
            # check all types
            from pandas.api.types import is_bool_dtype
            from pandas.api.types import is_string_dtype
            from pandas.api.types import is_numeric_dtype
            from pandas.api.types import is_float_dtype
            from pandas.api.types import is_integer_dtype
            from pandas.api.types import is_datetime64_any_dtype
            from pandas.api.types import is_categorical_dtype
            from pandas.api.types import is_object_dtype, is_list_like
            self.config['is_boolean'] = is_bool_dtype(in_type)
            self.config['is_string'] = is_string_dtype(in_type)
            self.config['is_numeric'] = is_numeric_dtype(in_type)
            self.config['is_integer'] = is_integer_dtype(in_type) or self.config['is_integer']
            self.config['is_float'] = is_float_dtype(in_type) if not self.config['is_integer'] else False
            self.config['is_datetime'] = is_datetime64_any_dtype(in_type)
            self.config['is_categorical'] = is_categorical_dtype(in_type)
            self.config['is_list'] = is_object_dtype(in_type) and is_list_like(in_type)

            if isinstance(in_type, ColumnSchema):
                self.config['is_categorical'] = in_type.is_categorical
                self.config['is_numeric'] = in_type.is_numeric or 'numeric' in str(in_type)
                self.config['is_categorical_string'] = in_type.is_categorical and not in_type.is_ordinal
                self.config['is_latlong'] = in_type.is_latlong

        self.post_fix()

    def post_fix(self):
        #post fix
        if 'is_datetime' in self.config and self.config['is_datetime']:
            self.config['is_text'] = False

    def copy_config_to(self, config):
        for k, v in self.config.items():
            if v:
                config[k] = v
        return config
    
    def copy_config_from(self, config):
        for k, v in config.items():
            if v is not False:
                self.config[k] = v
        self.post_fix()

    def mydump(self):
        return (self.name, list(k for k, v in self.config.items() if v is not False))

    def __repr__(self):
        return f"{self.mydump()}"
   
    @property
    def dtype_str(self):
        return str(dict((k, v) for k, v in self.config.items() if v))
         
    @property
    def is_boolean(self):
        return 'is_boolean' in self.config and self.config['is_boolean']

    @property
    def is_string(self):
        return 'is_string' in self.config and self.config['is_string']
    
    @property
    def is_numeric(self):
        return 'is_numeric' in self.config and self.config['is_numeric']
    
    @property
    def is_re_numeric(self):
        return 'is_re_numeric' in self.config and self.config['is_re_numeric']

    @property
    def is_float(self):
        return 'is_float' in self.config and self.config['is_float']

    @property
    def is_integer(self):
        return 'is_integer' in self.config and self.config['is_integer']

    @property
    def is_datetime(self):
        return 'is_datetime' in self.config and self.config['is_datetime']
    
    @property
    def is_categorical(self):
        return 'is_categorical' in self.config and self.config['is_categorical']
    
    @property
    def is_onehot(self):
        return 'is_onehot' in self.config and self.config['is_onehot'] is not False
    
    @property
    def is_list(self):
        return 'is_list' in self.config and self.config['is_list']
    
    @property
    def is_list_string(self):
        return 'is_list_string' in self.config and self.config['is_list_string'] is not False

    @property
    def is_categorical_and_string(self):
        if 'is_categorical_string' in self.config and self.config['is_categorical_string']:
            return True
        if 'is_categorical' in self.config and 'is_string' in self.config:
            return self.config['is_string'] and self.config['is_categorical']
        return False
    
    @property
    def is_coordinates(self):
        return 'is_latlong' in self.config and self.config['is_latlong']
    
    @property
    def is_encoded(self):
        return 'is_encoded' in self.config and self.config['is_encoded']

    @property
    def is_text(self):
        return 'is_text' in self.config and self.config["is_text"]
    

class DataFrameSchema(list):
    def __init__(self, df):
        for s_name in df.columns:
            s = df[s_name]
            super().append(SeriesSchema(s))

        
    


from .base import BaseOperation
import pandas as pd
from pyspark.sql import DataFrame as SparkDataFrame
from pyrecdp.core.utils import increment_encoder
import copy

class CategorifyOperation(BaseOperation):
    def __init__(self, op_base):
        super().__init__(op_base)
        self.feature_in_out = op_base.config
        self.support_spark_dataframe = False
        self.support_spark_rdd = False
    
    def get_function_pd(self):
        feature_in_out = copy.deepcopy(self.feature_in_out)
        def categorify(df):
            from sklearn.preprocessing import LabelEncoder
            import numpy
            import os
            for feature, (dict_path, feature_out) in feature_in_out.items():
                encoder = LabelEncoder()
                encoder.fit(df[feature])
                encoder = increment_encoder(encoder, dict_path)
                df[f"{feature_out}"] = pd.Series(encoder.transform(df[feature]))
                
            return df
        return categorify


from .base_api import base_api

class outbrain(base_api):
    def __init__(self):
        super().__init__()
        file_list = {
            'clicks': "clicks_train.csv",
            'documents_categories': "documents_categories.csv",
            'documents_entities': "documents_entities.csv",
            'documents_meta': "documents_meta.csv",
            'documents_topics': "documents_topics.csv",
            'events': "events.csv",
            'page_views': "page_views_sample.csv",
            'promoted_content': "promoted_content.csv"
        }

        self.saved_path = dict((f_name, self.download_url(f_path, f"https://outbrain-sampled.s3.us-west-2.amazonaws.com/{f_path}")) for f_name, f_path in file_list.items())

    def to_pandas(self, nrows = None):
        import pandas as pd
        return dict((f_name, pd.read_csv(f_path)) for f_name, f_path in self.saved_path.items())
         


import numpy as np
import pandas as pd
import copy
import os

import timeit

# `def pretty(d, indent=0):
#     from pyrecdp.primitives.operations import Operation
#     for key, value in d.items():
#         print('\t' * indent + str(key))
#         if isinstance(value, dict):
#             pretty(value, indent+1)
#         elif isinstance(value, list):
#             for a in value:
#                 pretty(a)
#         elif isinstance(value, Operation):
#             print('\t' * (indent+1) + str(value))

def callable_string_fix(func_str):
    func_str_lines = func_str.split('\n')
    import re
    bad_indent = "".join(re.findall("^(\s+)", func_str_lines[0]))
    if len(bad_indent) == 0:
        return func_str
    func_str = '\n'.join([i[len(bad_indent):] for i in func_str_lines])
    return func_str
    
def sequenced_union1d(a, b):
    if isinstance(a, np.ndarray):
        a_list = a.tolist()
    elif isinstance(a, list):
        a_list = a
    if isinstance(b, np.ndarray):
        b_list = b.tolist()
    elif isinstance(b, list):
        b_list = b
    a_dict = dict((i, 0) for i in a_list)
    [a_list.append(i) for i in b_list if i not in a_dict]
    return np.array(a_list)

def increment_encoder(encoder, dict_path):
    dirname = os.path.dirname(dict_path)
    if not os.path.exists(dirname):
        os.mkdir(dirname)
    if os.path.exists(dict_path):
        saved_dict = np.load(dict_path)
        # update dict
        ret = sequenced_union1d(encoder.classes_, saved_dict)
        encoder.classes_ = ret
    # save dictionary
    np.save(dict_path, encoder.classes_)
    
    return encoder
    
def deepcopy(dict_df):
    return copy.deepcopy(dict_df)
    
def get_sample_indices_pd(indices, target_num_rows):
    if isinstance(indices, pd.DataFrame):
        indices = indices.notna()
    if indices.size > target_num_rows:
        frac = target_num_rows/indices.size
        mask = pd.Series(np.random.choice(a=[False, True], size=(indices.size), p=[1-frac, frac]))
        indices = indices & mask
    return indices

def sample_read(file_path):
    if file_path.endswith('.csv'):
        return pd.read_csv(file_path, nrows = 10000)
    elif file_path.endswith('.parquet'):
        from pyarrow.parquet import ParquetFile
        import pyarrow as pa
        pf = ParquetFile(file_path) 
        sample_rows = next(pf.iter_batches(batch_size = 10000)) 
        return pa.Table.from_batches([sample_rows]).to_pandas() 
    else:
        raise NotImplementedError("now sample read only support csv and parquet")

def dump_fix(x):
    if isinstance(x, dict):
        for k, v in x.items():
            x[k] = dump_fix(v)
    elif isinstance(x, list) or isinstance(x, np.ndarray):
        for idx in range(len(x)):
            x[idx] = dump_fix(x[idx])
    elif isinstance(x, type):
        x = (x.__module__, x.__name__)
    elif isinstance(x, tuple):
        x = (dump_fix(x[0]), dump_fix(x[1]))
    elif hasattr(x, 'mydump'):
        x = x.mydump()
    elif callable(x):
        import inspect
        x = inspect.getsource(x)
    else:
        x = x
    return x

def class_name_fix(s):
    ret = s
    if isinstance(s, type):
        ret = s
    elif isinstance(s, tuple) or isinstance(s, list):
        import importlib
        module = importlib.import_module(s[0])
        ret = eval("module." + s[1])
    return ret
    
def is_text_series(s):
    from pandas.api import types as pdt
    import pandas as pd
    def try_text(s):
        if not pdt.is_string_dtype(s):
            return False
        if len(s) > 1000:
            # Sample to speed-up type inference
            result = s.sample(n=1000, random_state=0)
        try:
            avg_words = pd.Series(result).str.split().str.len().mean()
            if avg_words > 2:
                return True
        except:
            return False
        return False
    
    return try_text(s)

def is_tuple(s):
    t = type(s.loc[s.first_valid_index()]) if s.first_valid_index() >= 0 else None
    return isinstance(t, tuple)
       
def is_integer_convertable(s):
    from pandas.api.types import is_numeric_dtype
    if not is_numeric_dtype(s.dtype):
        return False
    s = s.fillna(0)
    if np.array_equal(s, s.astype(int)):
        return True
    return False
        
def is_encoded(s):
    line_id = s.first_valid_index()
    if line_id < 0:
        return False
    sample_data = s.loc[line_id:(line_id+1000)]
    from pyrecdp.primitives.generators.nlp import BertTokenizerDecode
    proc_ = BertTokenizerDecode().get_function()
    try:
        proc_(sample_data)
    except Exception as e:
        #print(e)
        return False
    return True
 
class Timer:
    level = 0
    viewer = None

    def __init__(self, name):
        self.name = name
        if Timer.viewer:
            Timer.viewer.display(f"{name} started ...")
        else:
            print(f"{name} started ...")

    def __enter__(self):
        self.start = timeit.default_timer()
        Timer.level += 1

    def __exit__(self, *a, **kw):
        Timer.level -= 1
        if Timer.viewer:
            Timer.viewer.display(
                f'{"  " * Timer.level}{self.name} took {timeit.default_timer() - self.start} sec')
        else:
            print(
                f'{"  " * Timer.level}{self.name} took {timeit.default_timer() - self.start} sec')

def update_linklist(linklist, key, value):
    if key not in linklist:
        linklist[key] = []
    linklist[key].append(value)
    return linklist


from .base import BaseOperation
import copy
from pyrecdp.core.utils import class_name_fix

class FeaturetoolsOperation(BaseOperation):
    def __init__(self, op_base):
        super().__init__(op_base)
        self.feature_in_out_map = op_base.config
        self.support_spark_dataframe = False
        self.support_spark_rdd = True

    def get_function_pd(self):
        feature_in_out_map = copy.deepcopy(self.feature_in_out_map)
        def generate_ft_feature(df):
            for in_feat_name, ops in feature_in_out_map.items():
                if in_feat_name in df.columns:
                    for op in ops:
                        op_object = class_name_fix(op[1])()
                        df[op[0]] = op_object(df[in_feat_name])
            return df
        return generate_ft_feature
    
    def get_function_spark(self, rdp):
        raise NotImplementedError(f"operations based on featuretools are not support Spark DataFrame yet.")


from .base_api import base_api

def try_finall(regex, input_str):
    import re
    ret = re.findall(regex, input_str)
    if len(ret) == 0:
        return None
    else:
        return ret[0]
    
def extract_info_from_report(report_content):
    method = None
    side = None
    p_id = None
    ret = {}
    lines = [i for i in report_content.split('\n') if i != '']
    for t in lines:
        if 'PATIENT NO.' in t:
            p_id = try_finall("\d+", t)
            method = None
            side = None                    
        elif 'SOFT TISSUE MAMMOGRAPHY REVEALED:' in t:
            method = 'DM'
        elif 'OPINION:' in t:
            method = 'OP'
        elif 'CONTRAST ENHANCED SPECTRAL MAMMOGRAPHY REVEALED:' in t:
            method = 'CESM'
        elif 'Right Breast' in t:
            side = "R"
        elif 'Left Breast' in t: 
            side = "L"
        else:
            if side is None or method is None:
                continue
            cur_key = f"{side}_{method}"
            if cur_key not in ret:
                ret[cur_key] = {}
                ret[cur_key]['Side'] = side
                ret[cur_key]['Patient_ID'] = int(p_id)
                ret[cur_key]['Type'] = method
                ret[cur_key]['symptoms'] = t
    
    return ret

class CESM_breast_cancer(base_api):
    def __init__(self, scale = 'full'):
        super().__init__()
        file_list = {
            'medical_report': "https://wiki.cancerimagingarchive.net/download/attachments/109379611/Medical%20reports%20for%20cases%20.zip?api=v2",
            'manual_annotations': "https://wiki.cancerimagingarchive.net/download/attachments/109379611/Radiology%20manual%20annotations.xlsx?api=v2"
        }

        self.saved_path = dict()
        self.saved_path['medical_report'] = self.download_url("Medical reports for cases", file_list['medical_report'], unzip = True)
        self.saved_path['manual_annotations'] = self.download_url("radiology_manual_annotations.xlsx", file_list['manual_annotations'])

    def to_pandas(self, nrows = None):
        import pandas as pd
        import os, docx2txt
        ret = {}
        ret['manual_annotations'] = pd.read_excel(self.saved_path['manual_annotations'], sheet_name="all")
        #ret['medical_report'] = {}
        medical_report_content = []
        medical_extracted = []
        for f in os.listdir(self.saved_path['medical_report']):
            try:
                file_content = docx2txt.process(os.path.join(self.saved_path['medical_report'], f))
            except Exception as e:
                file_content = ""
                Warning(e)
            medical_report_content.append(file_content)
            ext = extract_info_from_report(file_content)
            for side, line in ext.items():
                medical_extracted.append(line)

        ret['medical_report'] = pd.DataFrame.from_records(medical_extracted)
        
        return ret
         


from .base import BaseOperation
from pyrecdp.core import SeriesSchema
import pandas as pd
import copy

class TypeInferOperation(BaseOperation):
    def __init__(self, op_base):
        super().__init__(op_base)
        self.astype_feature_map = op_base.config
        self.support_spark_dataframe = False
        self.support_spark_rdd = True

    def get_function_pd(self):
        astype_feature_map = copy.deepcopy(self.astype_feature_map)
        def type_infer(df):
            for feature in astype_feature_map:
                feature_name, dest_type_list = feature[0], feature[1]
                # if 'is_datetime' in dest_type_list:
                #     df[feature_name] = pd.to_datetime(df[feature_name], errors='coerce', infer_datetime_format=True)
                if 'is_numeric' in dest_type_list:
                    df[feature_name] = pd.to_numeric(df[feature_name], errors='coerce')
            return df
        return type_infer
    
    def get_function_spark(self, rdp):
        raise NotImplementedError(f"TypeInferOperation spark dataframe is not supported yet.")


import ipywidgets as widgets
from IPython.display import display, clear_output

class BaseWidget:

    def __init__(self, display_flag = True):
        self.view = widgets.Output(layout={'border': '1px solid black'})
        if display_flag:
            display(self.view)

    def update(self, func, *args):
        with self.view:
            func(*args)

    def display(self, content):
        def display_func():
            if isinstance(content, str):
                print(content)
            else:
                display(content)
        self.update(display_func)

    def clear(self):
        self.view.clear_output()
    
    # def set_content(self, *embeddables, **kwargs):
    #     content_list = list(embeddables)
    #     self.outlet.set_content(widgets.VBox(content_list))

    # def set_title(self, title):
    #     self.outlet.set_title(title)

    # def hide(self):
    #     self.outlet.hide()

    # def is_displayed(self):
    #     return self.outlet.does_display(self)

    # def is_visible(self):
    #     return self.is_displayed() and self.outlet.is_visible()


import unittest
import sys
import pandas as pd
from pathlib import Path
pathlib = str(Path(__file__).parent.parent.resolve())
print(pathlib)
try:
    import pyrecdp
except:
    print("Not detect system installed pyrecdp, using local one")
    sys.path.append(pathlib)
from pyrecdp.autofe import RelationalBuilder

class TestRelationalBuilder(unittest.TestCase):
    def test_outbrain(self):
        train_data = {
            'clicks': "clicks_train.csv",
            'documents_categories': "documents_categories.csv",
            'documents_entities': "documents_entities.csv",
            'documents_meta': "documents_meta.csv",
            'documents_topics': "documents_topics.csv",
            'events': "events.csv",
            'page_views': "page_views_sample.csv",
            'promoted_content': "promoted_content.csv"}
        dir_path = f"{pathlib}/tests/data/outbrain/"
        train_data = dict((f_name, pd.read_csv(f"{dir_path}/{f_path}")) for f_name, f_path in train_data.items())
        pipeline = RelationalBuilder(dataset=train_data, label="clicked")
        #ret_df = pipeline.fit_transform(engine_type = 'pandas')
        # test with shape
        #self.assertEqual(ret_df.shape[0], 10000)
        #print(ret_df.dtypes)
        #self.assertTrue(ret_df.shape[1] >= 12)



from .base import BaseOperation
import copy

class RenameOperation(BaseOperation):
    def __init__(self, op_base):
        super().__init__(op_base)
        self.renamed = op_base.config
        self.support_spark_dataframe = True
        self.support_spark_rdd = True
        self.fast_without_dpp = True
        
    def get_function_pd(self):
        renamed = copy.deepcopy(self.renamed)
        def rename(df):
            return df.rename(columns = renamed)
        return rename

    def get_function_spark(self, rdp):
        def rename(df):
            for src, dst in self.renamed.items():
                df = df.withColumnRenamed(src, dst)
            
            return df
        return rename


from .base_api import base_api

class nyc_taxi(base_api):
    def __init__(self, scale = 'full'):
        super().__init__()
        self.scale = scale
        if scale == 'test':
            name = "test_nyc_taxi_fare.parquet"
            url = f"https://pyrecdp-testdata.s3.us-west-2.amazonaws.com/{name}"  
        elif scale == 'test_large':
            name = "nyc_taxi_fare_1M.csv"
            url = f"https://pyrecdp-testdata.s3.us-west-2.amazonaws.com/{name}"
        elif scale == 'full':
            name = "nyc_taxi_fare_cleaned.csv"
            url = "https://huggingface.co/datasets/Chendi/NYC_TAXI_FARE_CLEANED/resolve/main/nyc_taxi_fare_cleaned.csv"
                  
        self.saved_path = self.download_url(name, url)

    def to_pandas(self, nrows = None):
        import pandas as pd
        if self.scale == 'test':            
            return pd.read_parquet(self.saved_path)
        elif self.scale == 'test_large':
            return pd.read_csv(self.saved_path)
        elif self.scale == 'full':            
            if nrows:
                return pd.read_csv(self.saved_path, nrows = nrows)
            else:
                return pd.read_csv(self.saved_path)
         


from pyrecdp.datasets.download import *
from pyrecdp.datasets.nyc_taxi import *
from pyrecdp.datasets.twitter_recsys import *
from pyrecdp.datasets.amazon_product_review import *
from pyrecdp.datasets.outbrain import *
from pyrecdp.datasets.pretrained import *
from pyrecdp.datasets.ibm_fraud_detect import *
from pyrecdp.datasets.CESM_breast_cancer import *


from .base import BaseOperation
import copy

class FillNaOperation(BaseOperation):
    def __init__(self, op_base):
        super().__init__(op_base)
        self._fillna_feature_map = op_base.config
        self.support_spark_dataframe = True
        self.support_spark_rdd = True
    
    def get_function_pd(self):
        _fillna_feature_map = copy.deepcopy(self._fillna_feature_map)
        def fill_na(df):
            df.fillna(_fillna_feature_map, inplace=True, downcast=False)
            return df
        return fill_na

    def get_function_spark(self, rdp):
        def fill_na(df):
            return df.na.fill(self._fillna_feature_map)
        return fill_na


import json
import pandas as pd
from .dataframe import *
from pyspark.sql import DataFrame as SparkDataFrame
from pyspark import RDD as SparkRDD
from pyrecdp.core.utils import dump_fix
from IPython.display import display

class Operation:
    def __init__(self, idx, children, output, op, config):
        self.idx = idx
        self.children = children #input operation
        self.output = output #output schema
        self.op = op #func name
        self.config = config #operation config

    def __repr__(self):
        return repr(self.dump())
    
    def dump(self):
        dump_dict = {
            #'idx': self.idx,
            'children': self.children,
            #'output': self.output,
            'op': self.op,
            'config': dump_fix(self.config)
        }
        return dump_dict
    
    def instantiate(self):
        from .data import DataFrameOperation, DataLoader
        from .merge import MergeOperation
        from .name import RenameOperation
        from .category import CategorifyOperation
        from .drop import DropOperation
        from .fillna import FillNaOperation
        from .featuretools_adaptor import FeaturetoolsOperation
        from .geograph import HaversineOperation
        from .type import TypeInferOperation
        from .tuple import TupleOperation
        from .custom import CustomOperation
        from .encode import OnehotEncodeOperation, ListOnehotEncodeOperation, TargetEncodeOperation
        from pyrecdp.primitives.estimators.lightgbm import LightGBM

        operations_ = {
            'DataFrame': DataFrameOperation,
            'DataLoader': DataLoader,
            'merge': MergeOperation,
            'rename': RenameOperation,
            'categorify': CategorifyOperation,
            'drop': DropOperation,
            'fillna': FillNaOperation,
            'haversine': HaversineOperation,
            'tuple': TupleOperation,
            'type_infer': TypeInferOperation,
            'lightgbm': LightGBM,
            'onehot_encode': OnehotEncodeOperation,
            'list_onehot_encode': ListOnehotEncodeOperation,
            'target_encode': TargetEncodeOperation,
            'custom_operator': CustomOperation
        }

        if self.op in operations_:
            return operations_[self.op](self)
        else:
            try:
                return FeaturetoolsOperation(self)
            except:
                raise NotImplementedError(f"operation {self.op} is not implemented.")
 
    @staticmethod
    def load(idx, dump_dict):
        obj = Operation(idx, dump_dict['children'], None, dump_dict['op'], dump_dict['config'])
        return obj

class BaseOperation:
    def __init__(self, op_base):
        self.op = op_base
        self.cache = None
        self.support_spark_dataframe = False
        self.support_spark_rdd = False
        self.fast_without_dpp = False
       
    def __repr__(self) -> str:
        return self.op.op
        
    def execute_pd(self, pipeline, no_cache = False):
        _proc = self.get_function_pd()
        if not self.op.children or len(self.op.children) == 0:
            pass
        else:
            child_output = pipeline[self.op.children[0]].cache
            self.cache = _proc(child_output)
            
    def execute_spark(self, pipeline, rdp, no_cache = False):
        _convert = None
        if not self.op.children or len(self.op.children) == 0:
            pass
        else:
            child_output = pipeline[self.op.children[0]].cache
            if isinstance(child_output, SparkDataFrame):
                if self.support_spark_dataframe:
                    _proc = self.get_function_spark(rdp)
                elif self.support_spark_rdd:
                    _convert = SparkDataFrameToRDDConverter().get_function(rdp)
                    _proc = self.get_function_spark_rdd(rdp)
                else:
                    _convert = SparkDataFrameToDataFrameConverter().get_function(rdp)
                    _proc = self.get_function_pd()
            elif isinstance(child_output, SparkRDD):
                if self.support_spark_rdd:
                    _proc = self.get_function_spark_rdd(rdp)
                elif self.support_spark_dataframe:
                    _convert = RDDToSparkDataFrameConverter().get_function(rdp)
                    _proc = self.get_function_spark(rdp)
                else:
                    _convert = RDDToDataFrameConverter().get_function(rdp)
                    _proc = self.get_function_pd()
            elif isinstance(child_output, pd.DataFrame):
                if self.fast_without_dpp:
                    _proc = self.get_function_pd()
                elif self.support_spark_rdd:
                    _convert = DataFrameToRDDConverter().get_function(rdp)
                    _proc = self.get_function_spark_rdd(rdp)
                elif self.support_spark_dataframe:
                    _convert = DataFrameToSparkDataFrameConverter().get_function(rdp)
                    _proc = self.get_function_spark(rdp)
                else:
                    _proc = self.get_function_pd()
            else:
                raise ValueError(f"child cache is not recognized {child_output}")
        
            if _convert:
                child_output = _convert(child_output)
            self.cache = _proc(child_output)
            #print(self.cache.take(1))

    def get_function_spark_rdd(self, rdp):
        actual_func = self.get_function_pd()
        def transform(iter, *args):
            for x in iter:
                yield actual_func(x, *args)
        def base_spark_feature_generator(rdd):
            return rdd.mapPartitions(transform)
        return base_spark_feature_generator


from .base import BaseOperation

class TupleOperation(BaseOperation):        
    def __init__(self, op_base):
        super().__init__(op_base)
        self.support_spark_dataframe = False
        self.support_spark_rdd = True
        self.feature_in = op_base.config['src']
        self.feature_out = op_base.config['dst']

    def get_function_pd(self):
        feature_in = self.feature_in.copy()
        feature_out = self.feature_out
        def process(df):
            df[feature_out] = df[feature_in].apply(tuple, axis=1)
            return df
        return process

    def get_function_spark(self, rdp):
        raise NotImplementedError(f"CoordinatesOperation spark dataframe is not supported yet.")


from .base import BaseFeatureGenerator as super_class
from .featuretools_adaptor import FeaturetoolsBasedFeatureGenerator
from pyrecdp.core import SeriesSchema
from pyrecdp.primitives.operations import Operation
from featuretools.primitives import Haversine

class GeoFeatureGenerator(FeaturetoolsBasedFeatureGenerator):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.op_list = [Haversine]
        self.op_name = 'haversine'

    def fit_prepare(self, pipeline, children, max_idx):
        is_useful = False
        pa_schema = pipeline[children[0]].output
        for pa_field in pa_schema:
            if pa_field.is_coordinates:
                self.feature_in.append(pa_field.name)
                is_useful = True
        if len(self.feature_in) == 2:
            # we assume we can calculate distance between them
            out_feat_name = f"haversine_{'_'.join(self.feature_in)}"
            op = self.op_list[0]()
            out_feat_type = op.return_type
            out_schema = SeriesSchema(out_feat_name, out_feat_type)
            self.feature_in_out_map[str(self.feature_in)] = (out_schema.name, self.op_list[0])
            pa_schema.append(out_schema)

        if is_useful:
            cur_idx = max_idx + 1
            config = self.feature_in_out_map
            pipeline[cur_idx] = Operation(cur_idx, children, pa_schema, op = 'haversine', config = config)
            return pipeline, cur_idx, cur_idx
        else:
            return pipeline, children[0], max_idx


class Point:
    def __init__(self, prefix = None, longitude = None, latitude = None):
        self.feature_name_prefix = prefix
        self.longitude = longitude
        self.latitude = latitude
        
    def __eq__(self, other):
        return self.feature_name_prefix == other.feature_name_prefix
    
    def update(self, other):
        assert(self == other)
        self.longitude = other.longitude if other.longitude else self.longitude
        self.latitude = other.latitude if other.latitude else self.latitude

    def get_feature_name(self):
        return f"{self.feature_name_prefix}_coordinates"
        
    def get_feature_type(self):
        from woodwork.column_schema import ColumnSchema
        from woodwork.logical_types import LatLong
        return ColumnSchema(logical_type=LatLong)
    
    def get_config(self):
        return {'src':[self.latitude, self.longitude], 'dst': self.get_feature_name()}
 

class CoordinatesInferFeatureGenerator(super_class):        
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.points = []

    def fit_prepare(self, pipeline, children, max_idx):
        is_useful = False
        pa_schema = pipeline[children[0]].output
        for field in pa_schema:
            if self.is_coor_related_name_and_set(field.name):
                is_useful = True
        for p in self.points:
            out_schema = SeriesSchema(p.get_feature_name(), p.get_feature_type()) 
            pa_schema.append(out_schema)
        if is_useful:
            cur_idx = max_idx
            child = children[0]
            for p in self.points:
                cur_idx = cur_idx + 1
                config = p.get_config()
                pipeline[cur_idx] = Operation(cur_idx, [child], pa_schema, op = 'tuple', config = config)
                child = cur_idx
            return pipeline, cur_idx, cur_idx
        else:
            return pipeline, children[0], max_idx
    
    def is_coor_related_name_and_set(self, f_name):
            coor_related_names = ["longitude", "latitude", "coordinates", "point", "latlong", "longlat"]
            def get_prefix(in_name, remove_fix):
                return in_name.lower().replace(remove_fix, "").replace("_", "")
                
            for to_detect in coor_related_names:
                if to_detect in f_name.lower():
                    if to_detect == "longitude":
                        point = Point(longitude = f_name, prefix = get_prefix(f_name, to_detect))
                    elif to_detect == "latitude":
                        point = Point(latitude = f_name, prefix = get_prefix(f_name, to_detect))
                    
                    update_inline = False
                    for exist_point in self.points:
                        if exist_point == point:
                            exist_point.update(point)
                            update_inline = True
                    if not update_inline:
                        self.points.append(point)
                    return True
                    
            return False


from .base import BaseFeatureGenerator as super_class
from pyrecdp.core import SeriesSchema
from pyrecdp.core.utils import is_text_series
from pyrecdp.core.utils import Timer
import pandas as pd
import numpy as np

def draw_xy_scatter_plot(xy_scatter_features, feature_data, y, row_height, n_plot_per_row):
    from plotly.subplots import make_subplots
    import plotly.graph_objs as go
    import math
    
    n_feat = len(xy_scatter_features)
    n_row = math.ceil(n_feat / n_plot_per_row)
    n_col = n_plot_per_row if n_feat > n_plot_per_row else n_feat
    subplot_titles = xy_scatter_features

    fig_list = make_subplots(rows=n_row, cols=n_col, subplot_titles  = subplot_titles, y_title = y.name)

    indices = y.notna()
    if indices.size > 1000:
        frac = 1000/indices.size
        mask = pd.Series(np.random.choice(a=[False, True], size=(indices.size), p=[1-frac, frac]))
        indices = indices & mask

    for idx, c_name in enumerate(xy_scatter_features):
        feature = feature_data[c_name]
        # if string, wrap when too long
        sch = SeriesSchema(feature)
        is_feature_string = True if (sch.is_string or sch.is_categorical_and_string) else False
        
        if is_feature_string:
            tmp = feature[indices].str.slice(0, 12, 1)
            fig = go.Scatter(x=tmp, y=y[indices], mode='markers', name=c_name, showlegend=False)
        else:
            fig = go.Scatter(x=feature[indices], y=y[indices], mode='markers', name=c_name, showlegend=False)

        fig_list.add_trace(fig, row = int(idx / n_plot_per_row) + 1, col = ((idx % n_plot_per_row) + 1))

    fig_list.update_layout(height=row_height * n_row, width=400 * n_col)
    
    return fig_list

def draw_mapbox_plot(mapbox_scatter_features, feature_data):
    import plotly.graph_objs as go
    from shapely.geometry import MultiPoint
    fig_list = go.Figure()

    for c_name in mapbox_scatter_features:
        feature = feature_data[c_name]
        if len(feature) > 10000:
            feature = feature.sample(n=10000, random_state=123)
        coords = feature.to_list()
        lats = [c[0] for c in coords]
        lons = [c[1] for c in coords]
        fig_list.add_trace(
            go.Scattermapbox(
                lat=lats,
                lon=lons,
                name=c_name
            )
        )
        points = MultiPoint(coords)
    fig_list.update_layout(
        autosize=False,
        mapbox=dict(
            style='open-street-map',
            bearing=0,
            center=dict(
                lat=points.centroid.x,
                lon=points.centroid.y
            ),
            pitch=0,
            zoom=8
        ),
    )
        
    return fig_list
class StatisticsFeatureGenerator(super_class):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def fit_prepare(self, pipeline, children, max_idx):
        return pipeline, children[0], max_idx
    
    def update_feature_statistics(self, X, y):
        overview_info = {}
        overview_detail = {}
        overview_info['Number of Features'] = X.shape[1]
        overview_info['Number of Rows'] = X.shape[0]
        interactions_detail = self.get_interactive_plot(X, y)
        
        length = X.shape[0]
        for feature_name in X.columns:
            with Timer(f"prepare info for {feature_name}"):
                feature = X[feature_name]
                desc_info = dict((k, v) for k, v in feature.describe().to_dict().items() if k not in ['count'])
                if 'unique' in desc_info:
                    n_unique = desc_info['unique']
                else:
                    n_unique = feature.nunique()

                feature_type = SeriesSchema(feature).dtype_str
                feature_type = "text" if is_text_series(feature) else feature_type
                
                stat = {'type': feature_type, 'unique': {"u": n_unique, "m": length}, 'quantile':desc_info}
                if feature_name not in overview_detail:
                    overview_detail[feature_name] = stat
        
        data_stats = {}
        data_stats["overview"] = (overview_info, overview_detail)
        data_stats['interactions']=(dict(), interactions_detail)
        
        return data_stats
    
    def get_interactive_plot(self, feature_data, y):
        from plotly.subplots import make_subplots
        import plotly.graph_objs as go
        from plotly.offline import plot
        row_height = 300
        n_plot_per_row = 2
        
        # we will create types of plot
        xy_scatter_features = []
        mapbox_scatter_features = []
        word_cloud_features = []
        time_series_features = []
        
        for idx, c_name in enumerate(feature_data.columns):
            feature = feature_data[c_name]
            # if string, wrap when too long
            sch = SeriesSchema(feature)
            is_coord = True if sch.is_coordinates else False
            
            if is_coord:
                mapbox_scatter_features.append(c_name)
            else:
                xy_scatter_features.append(c_name)
        
        ret = {}
        ret = {"error": False}

        # draw xy scatter
        if len(xy_scatter_features) > 0:
            with Timer("Draw xy scatter plot"):
                fig_list = draw_xy_scatter_plot(xy_scatter_features, feature_data, y, row_height, n_plot_per_row)
            ret['html'] = plot(fig_list, output_type='div')
        
        # draw mapbox
        if len(mapbox_scatter_features) > 0:
            with Timer("Draw mapbox plot"):
                fig_list = draw_mapbox_plot(mapbox_scatter_features, feature_data)
            ret['html'] += plot(fig_list, output_type='div')
        
        return ret



from .base import BaseFeatureGenerator as super_class

class BinnedFeatureGenerator(super_class):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def fit_prepare(self, pipeline, children, max_idx):
        return pipeline, children[0], max_idx


from .binned import BinnedFeatureGenerator
from .category import CategoryFeatureGenerator
from .datetime import DatetimeFeatureGenerator
from .drop import DropUselessFeatureGenerator
from .name import RenameFeatureGenerator
from .fillna import FillNaFeatureGenerator
from .statics import StatisticsFeatureGenerator
from .type import TypeInferFeatureGenerator, TypeCheckFeatureGenerator,TypeConvertFeatureGenerator
from .nlp import DecodedTextFeatureGenerator, TextFeatureGenerator
from .geograph import GeoFeatureGenerator, CoordinatesInferFeatureGenerator
from .relation import RelationalFeatureGenerator
from .encode import OneHotFeatureGenerator, ListOneHotFeatureGenerator, TargetEncodeFeatureGenerator, LabelEncodeFeatureGenerator
from .feature_transform import ConvertToNumberFeatureGenerator

feature_infer_list = [
    TypeInferFeatureGenerator,   
]

relation_builder_list = [
    RelationalFeatureGenerator
]

label_feature_generator_list = [
    RenameFeatureGenerator,
    TypeConvertFeatureGenerator,
    FillNaFeatureGenerator,
    LabelEncodeFeatureGenerator,
]

pre_feature_generator_list = [
    CoordinatesInferFeatureGenerator,
    ConvertToNumberFeatureGenerator,
    TypeConvertFeatureGenerator,
    FillNaFeatureGenerator,
    RenameFeatureGenerator,
]

transformation_generator_list = [
    DecodedTextFeatureGenerator,
    DatetimeFeatureGenerator,
    GeoFeatureGenerator,
    TextFeatureGenerator,
]

local_encode_generator_list = [
    OneHotFeatureGenerator,
    ListOneHotFeatureGenerator,
]

pre_enocode_feature_generator_list = [
    DropUselessFeatureGenerator,
]

global_dict_index_generator_list = [
    BinnedFeatureGenerator,
    CategoryFeatureGenerator,
    #TargetEncodeFeatureGenerator
]

post_feature_generator_list = [
    RenameFeatureGenerator
]

final_generator_list = [
    TypeCheckFeatureGenerator,
    DropUselessFeatureGenerator,
]


import os
import re
from pyspark import *
from pyspark.sql import *
from py4j.protocol import Py4JJavaError
from py4j.java_gateway import JavaObject
from py4j.java_collections import ListConverter, JavaArray, JavaList, JavaMap
import pyspark.sql.types as spk_type
import pyspark.sql.functions as spk_func
import psutil

from pathlib import Path
pathlib = Path(__file__).parent.parent.resolve()

def create_spark_context():
    hname = os.uname()[1]
    paral = psutil.cpu_count(logical=False)
    total_mem = int((psutil.virtual_memory().total / (1 << 20)) * 0.8)
    print(f"Will assign {paral} cores and {total_mem} M memory for spark")
    spark = SparkSession.builder.master(f'local[{paral}]')\
                .appName("pyrecdp_spark_local")\
                .config("spark.executorEnv.HF_DATASETS_OFFLINE", "1")\
                .config("spark.executorEnv.TRANSFORMERS_OFFLINE", "1")\
                .config("spark.executorEnv.TF_CPP_MIN_LOG_LEVEL", "2")\
                .config("spark.executorEnv.PYTHONPATH", pathlib)\
                .config("spark.driver.memory", f"{total_mem}M")\
                .config("spark.driver.maxResultSize","64g")\
                .config("spark.sql.execution.arrow.pyspark.enabled","true")\
                .config("spark.sql.parquet.int96RebaseModeInWrite","CORRECTED")\
                .getOrCreate()
    spark.sparkContext.setLogLevel("ERROR")
    # try:
    #     spark = SparkSession.builder.master(f'spark://{hname}:7077')\
    #             .appName("pyrecdp_spark_standalone").getOrCreate()
    #     SparkContext.addPyFile() //this need to be zip, implement later
    # except:
    #     spark = SparkSession.builder.master(f'local[*]')\
    #             .appName("pyrecdp_spark_local").getOrCreate()
    return spark
    
def convert_to_spark_dict(orig_dict, schema=['dict_col', 'dict_col_id']):
    ret = []
    for row_k, row_v in orig_dict.items():
        ret.append({schema[0]: row_k, schema[1]: row_v})
    return ret


def convert_to_spark_df(orig_dict, spark):
    df = spark.createDataFrame(convert_to_spark_dict(orig_dict))
    df = df.withColumn('dict_col_id', spk_func.col('dict_col_id').cast(spk_type.IntegerType()))
    return df

def list_dir(path, only_get_one = True):
    source_path_dict = {}
    dirs = os.listdir(path)
    for files in dirs:
        try:
            sub_dirs = os.listdir(path + "/" + files)
            if not only_get_one:
                source_path_dict[files] = []
            for file_name in sub_dirs:
                if (file_name.endswith('parquet') or file_name.endswith('csv')):
                    if not only_get_one:
                        source_path_dict[files].append(os.path.join(
                            path, files, file_name))
                    else:
                        source_path_dict[files] = os.path.join(
                            path, files, file_name)
        except:
            if not only_get_one:
                source_path_dict[files] = [os.path.join(path, files)]
            else:
                source_path_dict[files] = os.path.join(path, files)
    return source_path_dict


def parse_size(size):
    units = {"B": 1, "KB": 2**10, "MB": 2**20, "GB": 2**30, "TB": 2**40, "K": 2**10, "M": 2**20, "G": 2**30, "T": 2**40}
    size = size.upper()
    number, unit = re.findall('(\d+)(\w*)', size)[0]
    return int(float(number)*units[unit])


def parse_cores_num(local_str):
    number = re.findall('local\[(\d+)\]', local_str)
    if len(number) > 0:
        return int(number[0])
    else:
        return os.cpu_count()


def get_estimate_size_of_dtype(dtype_name):
    units = {'byte': 1, 'short': 2, 'int': 4, 'long': 8, 'float': 4, 'double': 8, 'string': 10}
    return units[dtype_name] if dtype_name in units else 4


def get_dtype(df, colname):
    return [dtype for name, dtype in df.dtypes if name == colname][0]


def _j4py(spark, r):
    sc = spark.sparkContext
    if isinstance(r, JavaObject):
        clsName = r.getClass().getSimpleName()
        # convert RDD into JavaRDD
        if clsName != 'JavaRDD' and clsName.endswith("RDD"):
            r = r.toJavaRDD()
            clsName = 'JavaRDD'
        elif clsName == 'JavaRDD':
            jrdd = sc._jvm.SerDe.javaToPython(r)
            return RDD(jrdd, sc)
        elif clsName == 'DataFrame' or clsName == 'Dataset':
            return DataFrame(r, SQLContext.getOrCreate(sc))
        else:
             raise NotImplementedError(f"can't convert {clsName} {r} to java")
    else:
        raise NotImplementedError(f"can't convert {r} to java")
    return r


def _py4j(obj, gateway = None):
    """ Convert Python object into Java """
    if isinstance(obj, DataFrame):
        obj = obj._jdf
    elif isinstance(obj, SparkContext):
        obj = obj._jsc
    elif isinstance(obj, (list, tuple)):
        obj = ListConverter().convert([_py4j(x) for x in obj],
                                      gateway._gateway_client)
    elif isinstance(obj, JavaObject):
        pass
    elif isinstance(obj, (int, float, bool, bytes, str)):
        pass
    else:
        raise NotImplementedError(f"can't convert {str(obj)} to python")
    return obj    

def jvm_categorify(spark, df, col_name, dict_df):
    gateway = spark.sparkContext._gateway
    _jcls = gateway.jvm.org.apache.spark.sql.api.Categorify
    return _j4py(spark, _jcls.categorify(_py4j(spark.sparkContext), _py4j(df), _py4j(col_name), _py4j(dict_df)))

def jvm_categorify_for_array(spark, df, col_name, dict_df):
    gateway = spark.sparkContext._gateway
    _jcls = gateway.jvm.org.apache.spark.sql.api.CategorifyForArray
    return _j4py(spark, _jcls.categorify(_py4j(spark.sparkContext), _py4j(df), _py4j(col_name), _py4j(dict_df)))

def jvm_categorify_for_multi_level_array(spark, df, col_name, dict_df, sep_list):
    gateway = spark.sparkContext._gateway
    _jcls = gateway.jvm.org.apache.spark.sql.api.CategorifyForMultiLevelArray
    return _j4py(spark, _jcls.categorify(_py4j(spark.sparkContext), _py4j(df), _py4j(col_name), _py4j(dict_df), _py4j(sep_list, gateway)))

def jvm_categorify_by_freq_for_array(spark, df, col_name, dict_df):
    gateway = spark.sparkContext._gateway
    _jcls = gateway.jvm.org.apache.spark.sql.api.CategorifyByFreqForArray
    return _j4py(spark, _jcls.categorify(_py4j(spark.sparkContext), _py4j(df), _py4j(col_name), _py4j(dict_df)))

def jvm_get_negative_sample(spark, df, col_name, dict_df, neg_cnt):
    gateway = spark.sparkContext._gateway
    _jcls = gateway.jvm.org.apache.spark.sql.api.NegativeSample
    return _j4py(spark, _jcls.add(_py4j(spark.sparkContext), _py4j(df), _py4j(col_name), _py4j(dict_df), _py4j(neg_cnt)))

def jvm_get_negative_feature(spark, df, tgt_name, src_name, dict_df, neg_cnt):
    gateway = spark.sparkContext._gateway
    _jcls = gateway.jvm.org.apache.spark.sql.api.NegativeFeature
    return _j4py(spark, _jcls.add(_py4j(spark.sparkContext), _py4j(df), _py4j(tgt_name), _py4j(src_name), _py4j(dict_df), _py4j(neg_cnt)))

def jvm_get_negative_feature_for_array(spark, df, tgt_name, src_name, dict_df, neg_cnt):
    gateway = spark.sparkContext._gateway
    _jcls = gateway.jvm.org.apache.spark.sql.api.NegativeFeatureForArray
    return _j4py(spark, _jcls.add(_py4j(spark.sparkContext), _py4j(df), _py4j(tgt_name), _py4j(src_name), _py4j(dict_df), _py4j(neg_cnt)))

def jvm_scala_df_test_int(spark, df, col_name):
    gateway = spark.sparkContext._gateway
    _jcls = gateway.jvm.org.apache.spark.sql.api.ScalaDFTest
    return _j4py(spark, _jcls.scala_test_int(_py4j(spark.sparkContext), _py4j(df), _py4j(col_name)))

def jvm_scala_df_test_str(spark, df, col_name):
    gateway = spark.sparkContext._gateway
    _jcls = gateway.jvm.org.apache.spark.sql.api.ScalaDFTest
    return _j4py(spark, _jcls.scala_test_str(_py4j(spark.sparkContext), _py4j(df), _py4j(col_name)))

def jvm_scala_df_test_broadcast(spark, df, col_name, dict_df):
    gateway = spark.sparkContext._gateway
    _jcls = gateway.jvm.org.apache.spark.sql.api.ScalaDFTest
    return _j4py(spark, _jcls.scala_test_broadcast(_py4j(spark.sparkContext), _py4j(df), _py4j(col_name), _py4j(dict_df)))



from .base import BaseFeatureGenerator as super_class
from pyrecdp.core import SeriesSchema
from pyrecdp.primitives.operations import Operation
import pandas as pd

def get_default_value(at: SeriesSchema):
    if at.is_boolean:
        return False
    elif at.is_numeric:
        return -1
    elif at.is_datetime:
        return 0
    elif at.is_string:
        return ""
    return None

class FillNaFeatureGenerator(super_class):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def fit_prepare(self, pipeline, children, max_idx):
        pa_schema = pipeline[children[0]].output
        self._fillna_feature_map = {}
        for field in pa_schema:
            default_value = get_default_value(field)
            if default_value != None:
                self._fillna_feature_map[field.name] = default_value
        if len(self._fillna_feature_map) > 0:
            cur_idx = max_idx + 1
            config = self._fillna_feature_map
            pipeline[cur_idx] = Operation(cur_idx, children, pa_schema, op = 'fillna', config = config)
            return pipeline, cur_idx, cur_idx
        else:
            return pipeline, children[0], max_idx




from pyrecdp.widgets import BaseWidget

class TableViewWidget(BaseWidget):

    def __init__(self, *args, parent_tabs=None, **kwargs):
        super().__init__(*args, **kwargs)
        self.parent_tabs = parent_tabs
        self._df_is_outdated = False

    def df_did_change(self):
        """
        Call this method when the DataFrame did change
        """
        self._df_is_outdated = True

    def tab_got_selected(self):
        """
        Call this method when the tab of this TabViewable got selected
        """
        if self._df_is_outdated:
            self._df_is_outdated = False
            self.render()

    def render_in(self, outlet):
        if isinstance(outlet, Window):
            super().render_in(outlet)
        else:
            # the outlet is a TabSection
            # this is a hotfix to enable the same API for TabViewables e.g.
            # `TabViewable().render_in(tab_section)`
            outlet.add_tab(self)

    def set_content(self, *embeddables, **kwargs):
        # Attention: this method was overriden and copied from Viewable
        # because we need to add a class to the wrapping widgets.VBox
        # This ensures that TabViewables have a min-height
        content_list = list(embeddables)
        box = widgets.VBox(content_list)
        box.add_class("bamboolib-min-height-for-tab-viewables")
        self.outlet.set_content(box)

    def _ipython_display_(self, *args, **kwargs):
        """
        This method is called by Jupyter when trying to display the object.
        We are setting up all the surrounding classes that are needed for the experience
        of displaying a TabViewable within a properly setup TabSection
        """
        from bamboolib.wrangler import (
            Wrangler,
        )  # inline import to prevent circular import

        df_manager = self.df_manager
        tab_section = TabSection(df_manager)
        tab_section.add_tab(
            Wrangler(df_manager=df_manager, parent_tabs=tab_section), closable=False
        )
        tab_section.add_tab(self, closable=False)
        display(tab_section)


import unittest
import sys
import pandas as pd
from pathlib import Path
pathlib = str(Path(__file__).parent.parent.resolve())
print(pathlib)
try:
    import pyrecdp
except:
    print("Not detect system installed pyrecdp, using local one")
    sys.path.append(pathlib)
from pyrecdp.autofe import FeatureWrangler
from IPython.display import display


class TestFeatureWranglerPandasBased(unittest.TestCase):

    def test_nyc_taxi(self):
        train_data = pd.read_parquet(f"{pathlib}/tests/data/test_nyc_taxi_fare.parquet")
        pipeline = FeatureWrangler(dataset=train_data, label="fare_amount")
        ret_df = pipeline.fit_transform(engine_type = 'pandas')
        # test with shape
        self.assertEqual(ret_df.shape[0], 10000)
        self.assertTrue(ret_df.shape[1] >= 12)
        
    def test_twitter_recsys(self):
        train_data = pd.read_parquet(f"{pathlib}/tests/data/test_twitter_recsys.parquet")
        pipeline = FeatureWrangler(dataset=train_data, label="reply")
        ret_df = pipeline.fit_transform(engine_type = 'pandas')
        # test with shape
        self.assertEqual(ret_df.shape[0], 10000)
        self.assertTrue(ret_df.shape[1] >= 31)
    
    def test_amazon(self):
        train_data = pd.read_table(f"{pathlib}/tests/data/amazon_reviews_us_Books.tsv", on_bad_lines='skip')
        pipeline = FeatureWrangler(dataset=train_data, label="star_rating")
        ret_df = pipeline.fit_transform(engine_type = 'pandas')
        # test with shape
        self.assertEqual(ret_df.shape[0], 9667)
        self.assertTrue(ret_df.shape[1] >= 16)
        
    def test_frauddetect(self):
        from pyrecdp.datasets import ibm_fraud_detect
        train_data = ibm_fraud_detect().to_pandas('test')
        pipeline = FeatureWrangler(dataset=train_data, label="Is Fraud?")
        ret_df = pipeline.fit_transform(engine_type = 'pandas')
        display(ret_df)
        
class TestFeatureWranglerSparkBased(unittest.TestCase):

    def test_nyc_taxi(self):
        train_data = pd.read_parquet(f"{pathlib}/tests/data/test_nyc_taxi_fare.parquet")
        pipeline = FeatureWrangler(dataset=train_data, label="fare_amount")
        ret_df = pipeline.fit_transform(engine_type = 'spark')
        # test with shape
        self.assertEqual(ret_df.shape[0], 10000)
        self.assertTrue(ret_df.shape[1] >= 12)
        
    def test_twitter_recsys(self):
        train_data = pd.read_parquet(f"{pathlib}/tests/data/test_twitter_recsys.parquet")
        pipeline = FeatureWrangler(dataset=train_data, label="reply")
        ret_df = pipeline.fit_transform(engine_type = 'spark')
        # test with shape
        self.assertEqual(ret_df.shape[0], 10000)
        self.assertTrue(ret_df.shape[1] >= 31)

    def test_amazon(self):
        train_data = pd.read_table(f"{pathlib}/tests/data/amazon_reviews_us_Books.tsv", on_bad_lines='skip')
        pipeline = FeatureWrangler(dataset=train_data, label="star_rating")
        ret_df = pipeline.fit_transform(engine_type = 'spark')
        # test with shape
        self.assertEqual(ret_df.shape[0], 9667)
        self.assertTrue(ret_df.shape[1] >= 16)

    def test_frauddetect(self):
        from pyrecdp.datasets import ibm_fraud_detect
        train_data = ibm_fraud_detect().to_pandas('test')
        pipeline = FeatureWrangler(dataset=train_data, label="Is Fraud?")
        ret_df = pipeline.fit_transform(engine_type = 'spark')
        display(ret_df)



from .base_api import base_api

class pretrained(base_api):
    def __init__(self):
        super().__init__()
   
    def download(self, model_name):
        if model_name == "nyc_taxi_fare":
            name = "lightgbm_regression_nyc_taxi_fare_amount.mdl"
            url = f"https://pyrecdp-testdata.s3.us-west-2.amazonaws.com/{name}"            
            return self.download_url(name, url)
         


import uuid
from pyrecdp.utils import *
from pyspark.ml.feature import *
import pandas as pd
import numpy as np
import pyspark.sql.types as spk_type
import pyspark.sql.types as t
import pyspark.sql.functions as spk_func
import pyspark.sql.functions as f
from pyspark.sql import *
from pyspark import *
import os
import sys
from timeit import default_timer as timer
import logging
import shutil


class Encoder:
    def __init__(self, proc):
        self.op_name = "Encoder"
        self.uuid = proc.uuid
        self.tmp_id = proc.tmp_id
        self.path_prefix = proc.path_prefix
        self.current_path = proc.current_path
        self.dicts_path = proc.dicts_path
        self.spark = proc.spark
        self.tmp_materialzed_list = []

    def transform(self, train, valid, train_only=True):
        raise NotImplementedError("This is base Encoder class")

    def materialize(self, df, df_name="materialized_tmp"):
        tmp_id = self.tmp_id
        self.tmp_id += 1
        save_path = ""
        if df_name == "materialized_tmp":
            save_path = "%s/%s/tmp/%s-%s-%d" % (
                self.path_prefix, self.current_path, df_name, self.uuid, tmp_id)
            self.tmp_materialzed_list.append(save_path)
        else:
            save_path = "%s/%s/%s" % (self.path_prefix,
                                      self.current_path, df_name)
        df.write.format('parquet').mode('overwrite').save(save_path)
        return self.spark.read.parquet(save_path)


class TargetEncoder(Encoder):
    def __init__(self, proc, x_col_list, y_col_list, out_col_list, out_name, out_dtype=None, y_mean_list=None, smooth=20, seed=42,threshold=0):
        super().__init__(proc)
        self.op_name = "TargetEncoder"
        self.x_col_list = x_col_list
        self.y_col_list = y_col_list
        self.out_col_list = out_col_list
        self.out_dtype = out_dtype
        self.out_name = out_name
        self.y_mean_list = y_mean_list
        self.seed = seed
        self.smooth = smooth
        self.expected_list_size = len(y_col_list)
        self.threshold = threshold
        if len(self.out_col_list) < self.expected_list_size:
            raise ValueError("TargetEncoder __init__, input out_col_list should be same size as y_col_list")      
        if y_mean_list != None and len(self.y_mean_list) < self.expected_list_size:
            raise ValueError("TargetEncoder __init__, input y_mean_list should be same size as y_col_list")        

    def transform(self, df):
        x_col = self.x_col_list
        cols = ['fold', x_col] if isinstance(x_col, str) else ['fold'] + x_col
        agg_per_fold = df.groupBy(cols)
        agg_all = df.groupBy(x_col)

        per_fold_list = []
        all_list = []

        for i in range(0, self.expected_list_size):
            y_col = self.y_col_list[i]
            per_fold_list.append(f.count(y_col).alias(f'count_{y_col}'))
            per_fold_list.append(f.sum(y_col).alias(f'sum_{y_col}'))
            all_list.append(f.count(y_col).alias(f'count_all_{y_col}'))
            all_list.append(f.sum(y_col).alias(f'sum_all_{y_col}'))

        agg_per_fold = agg_per_fold.agg(*per_fold_list)
        agg_all = agg_all.agg(*all_list)
        agg_per_fold = agg_per_fold.join(agg_all, x_col, 'left')

        if self.threshold > 0:
            agg_all = agg_all.where(f.col(f'count_all_{self.y_col_list[0]}')>self.threshold)

        for i in range(0, self.expected_list_size):
            y_col = self.y_col_list[i]
            out_col = self.out_col_list[i]
            out_dtype = self.out_dtype
            y_mean = self.y_mean_list[i] if self.y_mean_list != None else None
            
            if y_mean is None:
                y_mean = np.array(df.groupBy().mean(y_col).collect())[0][0]
            mean = float(y_mean)
            smooth = self.smooth

            # print(agg_per_fold.dtypes)

            # prepare for agg_per_fold
            agg_per_fold = agg_per_fold.withColumn(
                f'count_all_{y_col}', f.col(f'count_all_{y_col}')-f.col(f'count_{y_col}'))
            agg_per_fold = agg_per_fold.withColumn(
                f'sum_all_{y_col}', f.col(f'sum_all_{y_col}')-f.col(f'sum_{y_col}'))
            agg_per_fold = agg_per_fold.withColumn(
                out_col,
                (f.col(f'sum_all_{y_col}') + f.lit(mean) * f.lit(smooth))/(f.col(f'count_all_{y_col}') + f.lit(smooth)))
            agg_all = agg_all.withColumn(
                out_col,
                (f.col(f'sum_all_{y_col}') + f.lit(mean) * f.lit(smooth))/(f.col(f'count_all_{y_col}') + f.lit(smooth)))
            if out_dtype is not None:
                agg_per_fold = agg_per_fold.withColumn(
                    out_col, f.col(out_col).cast(out_dtype))
                agg_all = agg_all.withColumn(
                    out_col, f.col(out_col).cast(out_dtype))
            agg_per_fold = agg_per_fold.drop(
                f'count_all_{y_col}', f'count_{y_col}', f'sum_all_{y_col}', f'sum_{y_col}')
            agg_all = agg_all.drop(f'count_all_{y_col}', f'sum_all_{y_col}')
        return (self.materialize(agg_per_fold, "%s/train/%s" % (self.dicts_path, self.out_name)),
                self.materialize(agg_all, "%s/test/%s" % (self.dicts_path, self.out_name)))


class CountEncoder(Encoder):
    def __init__(self, proc, x_col_list, y_col_list, out_col_list, out_name, train_generate=True):
        super().__init__(proc)
        self.op_name = "CountEncoder"
        self.x_col_list = x_col_list
        self.y_col_list = y_col_list
        self.out_col_list = out_col_list
        self.out_name = out_name        
        self.expected_list_size = len(y_col_list)
        self.train_generate = train_generate
        if len(self.out_col_list) < self.expected_list_size:
            raise ValueError("CountEncoder __init__, input out_col_list should be same size as y_col_list")

    def transform(self, df):
        x_col = self.x_col_list
        cols = [x_col] if isinstance(x_col, str) else x_col
        agg_all = df.groupby(cols)

        all_list = []

        for i in range(0, self.expected_list_size):
            y_col = self.y_col_list[i]
            out_col = self.out_col_list[i]
            all_list.append(f.count(y_col).alias(f'{out_col}'))

        agg_all = agg_all.agg(*all_list)
        for i in range(0, self.expected_list_size):
            out_col = self.out_col_list[i]
            agg_all = agg_all.withColumn(out_col, f.col(out_col).cast(spk_type.IntegerType()))

        if self.train_generate:
            return (self.materialize(agg_all, "%s/train/%s" % (self.dicts_path, self.out_name)),
                    self.materialize(agg_all, "%s/test/%s" % (self.dicts_path, self.out_name)))
        else:
            return (self.materialize(agg_all, "%s/test/%s" % (self.dicts_path, self.out_name)))



class FrequencyEncoder(Encoder):
    def __init__(self, proc, x_col_list, y_col_list, out_col_list, out_name):
        super().__init__(proc)
        self.op_name = "FrequencyEncoder"
        self.x_col_list = x_col_list
        self.y_col_list = y_col_list
        self.out_col_list = out_col_list
        self.out_name = out_name        
        self.expected_list_size = len(y_col_list)
        if len(self.out_col_list) < self.expected_list_size:
            raise ValueError("FrequencyEncoder __init__, input out_col_list should be same size as y_col_list")

    def transform(self, df):
        length_df = df.count()
        x_col = self.x_col_list
        cols = [x_col] if isinstance(x_col, str) else x_col
        agg_all = df.groupby(cols)

        all_list = []

        for i in range(0, self.expected_list_size):
            y_col = self.y_col_list[i]
            out_col = self.out_col_list[i]
            all_list.append(f.count(y_col).alias(out_col))
        agg_all = agg_all.agg(*all_list)

        for i in range(0, self.expected_list_size):
            out_col = self.out_col_list[i]
            agg_all = agg_all.withColumn(out_col, (f.col(out_col)*1.0/length_df).cast(spk_type.FloatType()))
        return (self.materialize(agg_all, "%s/train/%s" % (self.dicts_path, self.out_name)),
                None)


from .base import BaseFeatureGenerator as super_class
from pyrecdp.primitives.operations import Operation
    
class RenameFeatureGenerator(super_class):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.renamed = {}

    def fit_prepare(self, pipeline, children, max_idx):
        pa_schema = pipeline[children[0]].output
        is_useful = False
        for pa_field in pa_schema:
            feature_name = pa_field.name
            if '.' in feature_name:
                feature_name = feature_name.replace('.', '__')
                is_useful = True
            if ' ' in feature_name:
                feature_name = feature_name.replace(' ', '_')
                is_useful = True
            if '?' in feature_name:
                feature_name = feature_name.replace('?', '')
                is_useful = True
            if is_useful:
                self.renamed[pa_field.name] = feature_name
        ret_schema = []
        for pa_field in pa_schema:
            if pa_field.name in self.renamed:
                pa_field.name = self.renamed[pa_field.name]
            ret_schema.append(pa_field)

        if is_useful:
            cur_idx = max_idx + 1
            config = self.renamed
            pipeline[cur_idx] = Operation(cur_idx, children, ret_schema, op = 'rename', config = config)
            return pipeline, cur_idx, cur_idx
        else:
            return pipeline, children[0], max_idx


import logging
from pyrecdp.core.utils import Timer

from pyrecdp.widgets import BaseWidget, TabWidget
from pyrecdp.autofe import FeatureProfiler, FeatureWrangler

import os

logging.basicConfig(format='%(asctime)s %(levelname)s:%(message)s', level=logging.ERROR, datefmt='%I:%M:%S')
logger = logging.getLogger(__name__)

class AutoFE:
    def __init__(self, dataset, label, only_pipeline = False, enable_profiler = True, *args, **kwargs):
        self.label = label
        self.data = dataset

        # prepare main view
        self.original_df_view = BaseWidget(display_flag=False)
        self.log_view = BaseWidget(display_flag=False)
        self.main_view = TabWidget([('log', self.log_view), ('original_df', self.original_df_view)])
        Timer.viewer = self.log_view

        # prepare pipeline        
        self.feature_generator = FeatureWrangler(self.data, self.data[self.label])

        # start data analysis and transform
        with Timer("AutoFE Data Wrangling"):
            # profile original dataframe
            self.original_df_view.display(self.data)
            with Timer("Analysis Original DataFrame"):
                self.feature_generator.fit_analyze()
                self.log_view.display(f"feature engineering pipeline after analysis:\n{self.get_transform_pipeline()}")

            # auto feature engineering
            with Timer("Auto Feature Engineering on dataset"):
                self.transformed_feature = self.feature_generator.fit_transform()

            # create view for transformed data
            self.transformed_df_view = BaseWidget(display_flag=False)
            self.main_view.append('transformed_df', self.transformed_df_view)
            self.transformed_df_view.display(self.get_transformed_data())

            # profile transformed dataframe
            with Timer("profiling transformed dataset"):
                self.transformed_df_profile = FeatureProfiler(self.transformed_feature, self.data[self.label]).visualize_analyze().show()
            

    def get_transform_pipeline(self):
        return "\n".join([f"Stage {i}: {[g.__class__ for g in stage]}" for i, stage in enumerate(self.feature_generator.generators)])
    
    def exclude_target(self, df):
        label = self.label if isinstance(self.label, list) else [self.label]
        feat_columns = [n for n in df.columns if n not in label]
        return df[feat_columns]
    
    def get_origin_feature_list(self):
        return self.exclude_target(self.data).dtypes

    def get_feature_list(self):
        return self.exclude_target(self.transformed_feature).dtypes
    
    def get_transformed_data(self):
        return self.transformed_feature
    
    def get_original_data(self):
        return self.data

    def get_problem_type(self):
        return self.problem_type



from pyrecdp.autofe.FeatureProfiler import *
from pyrecdp.autofe.FeatureWrangler import *
from pyrecdp.autofe.RelationalBuilder import *
from pyrecdp.autofe.DataEstimator import *
from pyrecdp.autofe.AutoFE import *


from jinja2 import Environment, PackageLoader
from pyrecdp.primitives.generators import *
from .BasePipeline import BasePipeline
from bokeh.resources import INLINE
ENV_LOADER = Environment(
    loader=PackageLoader("pyrecdp", "widgets/templates"),
)

CELL_HEIGHT_OVERRIDE = """<style>
                            div.output_scroll {
                              height: 850px;
                            }
                            div.cell-output>div:first-of-type {
                              max-height: 850px !important;
                            }
                          </style>"""

class FeatureVisulizer:
    def __init__(self, stats):    

        template_base = ENV_LOADER.get_template("base.html")
        context = {
            "resources": INLINE.render(),
            "title": "FeatureProfiler",
            "components": self.format_report(stats),
        }
        self.report = template_base.render(context=context)
    
    def format_report(self, stats):
        return stats
    
    def show(self):
        from IPython.display import (  # pylint: disable=import-outside-toplevel
                HTML,
                display,
            )

        display(HTML(self._repr_html_()))

    def _repr_html_(self) -> str:
        """
        Display report inside a notebook
        """
        return f"{CELL_HEIGHT_OVERRIDE}<div style='background-color: #fff;'>{self.report}</div>"
class FeatureProfiler(BasePipeline):        
    def __init__(self, dataset, label, *args, **kwargs):
        super().__init__(dataset, label)

        self.data_stats = None
        self._processed_data = self.feature_data

        self.generators.append([cls() for cls in feature_infer_list])

        self.fit_analyze()
    
    def visualize_analyze(self, engine_type = 'pandas', display = True):
        if not self.data_stats:
            feature_data = self.fit_transform(engine_type)
            self.data_stats = StatisticsFeatureGenerator().update_feature_statistics(feature_data, self.y)
            self._processed_data = feature_data
        return FeatureVisulizer(self.data_stats)

    def _debug_get_processed_data(self):
        return self._processed_data


from .base import BaseEstimator
import pandas as pd
import lightgbm as lgbm

class LightGBM(BaseEstimator):
    def __init__(self, op_base):
        super().__init__(op_base)
        self.support_spark_dataframe = False
        self.support_spark_rdd = False

    def get_func_train(self):
        label = self.config['label']
        metrics = self.config['metrics']
        objective = self.config['objective']
        model_file = self.config['model_file']
        train_test_splitter = self.get_splitter_func(self.config['train_test_splitter'])
        def train(df):
            train_sample, test_sample = train_test_splitter(df)
            
            x_train = train_sample.drop(columns=[label])
            y_train = train_sample[label].values

            x_val = test_sample.drop(columns=[label])
            y_val = test_sample[label].values

            lgbm_train = lgbm.Dataset(x_train, y_train, silent=False)
            lgbm_val = lgbm.Dataset(x_val, y_val, silent=False)
            
            params = {
                'boosting_type':'gbdt',
                'objective': objective,
                'nthread': 4,
                'num_leaves': 31,
                'learning_rate': 0.05,
                'max_depth': -1,
                'subsample': 0.8,
                'bagging_fraction' : 1,
                'max_bin' : 5000 ,
                'bagging_freq': 20,
                'colsample_bytree': 0.6,
                'metric': metrics,
                'min_split_gain': 0.5,
                'min_child_weight': 1,
                'min_child_samples': 10,
                'scale_pos_weight':1,
                'zero_as_missing': True,
                'seed':0,
                'num_rounds':2000,
                'num_boost_round': 2000,
                'early_stopping_rounds': 50
            }
            model = lgbm.train(params=params, train_set=lgbm_train, valid_sets=lgbm_val, verbose_eval=100)
            model.save_model(model_file, num_iteration=model.best_iteration) 
            return model
        return train
        
    def get_func_predict(self):
        label = self.config['label']
        metrics = self.config['metrics']
        model_file = self.config['model_file']
        evalute_func = self.get_evaluate_func(metrics)
        model = lgbm.Booster(model_file = model_file)
        def predict(test_df):
            contain_label = False
            if label in test_df.columns:
                x_val = test_df.drop(columns=[label])
                contain_label = True
            pred = model.predict(x_val)
            if contain_label:
                y_val = test_df[label].values
                score = evalute_func(y_val, pred)
                print(f"This test_data contains label, {metrics} is {score}")

            test_df['predict_value'] = pd.Series(pred)
            return test_df
    
        return predict


import pandas as pd

class PandasDataFrame(pd.DataFrame):
    def __init__(self, baseObject):
        self.__setattr__(self, '_wrapped_obj', baseObject)
        self.__setattr__(self, '_wrapped_obj', baseObject)

        #self._wrapped_obj = baseObject
        #self.sample_indices = None
        

    def may_sample(self):
        if self._wrapped_obj.shape[0] > 10000:
            return PandasDataFrame(self._wrapped_obj.sample(n=10000, random_state = 123))
        else:
            return self
    
    def __getitem__(self, key):
        if isinstance(key, list):
            return PandasDataFrame(self._wrapped_obj[key])
        else:
            return self._wrapped_obj[key]

class DataFrameAPI:
    @staticmethod
    def instiate(df):
        if isinstance(df, pd.DataFrame):
            from pandas_flavor import register_dataframe_method
            @register_dataframe_method
            def may_sample(df):
                if df.shape[0] > 100000:
                    return df.sample(n=100000, random_state = 123)
                else:
                    return df
            return df
        elif isinstance(df, str):
            # open the path depends on the size
            raise NotImplementedError("Instiate DataFrame based on path is not implemented yet.")
        raise NotImplementedError("Instiate DataFrame based on non-pandas dataframe is not implemented yet.")


from .base import BaseFeatureGenerator as super_class
from pyrecdp.core import SeriesSchema
from pyrecdp.primitives.operations import Operation

class FeaturetoolsBasedFeatureGenerator(super_class):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.feature_in = []
        self.feature_in_out_map = {}
        self.op_name = 'featuretools'
    
    def fit_prepare(self, pipeline, children, max_idx):
        is_useful = False
        pa_schema = pipeline[children[0]].output
        for in_feat_name in self.feature_in:
            is_useful = True
            self.feature_in_out_map[in_feat_name] = []
            for op in self.op_list:
                op_clz = op
                op = op_clz()
                out_feat_name = f"{in_feat_name}__{op.name}"
                out_feat_type = op.return_type
                out_schema = SeriesSchema(out_feat_name, out_feat_type)
                pa_schema.append(out_schema)
                self.feature_in_out_map[in_feat_name].append((out_schema.name, op_clz))
        if is_useful:
            cur_idx = max_idx + 1
            pipeline[cur_idx] = Operation(cur_idx, children, pa_schema, op = self.op_name, config = self.feature_in_out_map)
            return pipeline, cur_idx, cur_idx
        else:
            return pipeline, children[0], max_idx